unit appgraphic;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, IntfGraphics, Graphics, ExtCtrls, types, Dialogs, math, translation;

type
  TRGBAColor = record
    Red: Byte;
    Green: Byte;
    Blue: Byte;
    Alpha: Byte;
  end;
  PRGBAColor = ^TRGBAColor;
  TRGBAColorLine = array[0..32767] of TRGBAColor;
  PRGBAColorLine = ^TRGBAColorLine;

  TRGBColor = record
    Red: Byte;
    Green: Byte;
    Blue: Byte;
  end;
  PRGBColor = ^TRGBColor;
  TRGBColorLine = array[0..32767] of TRGBColor;
  PRGBColorLine = ^TRGBColorLine;

  { TGraphicItem }

  TGraphicItem = class
    private
      Images: array of TFPImageBitmap;
      ImageName: string;
      OriginX, OriginY: Integer;
      function GetImage(Pos: Cardinal): TFPImageBitmap;
      function GetImagesCount: Integer;
    public
      constructor Create(FileName, Name, FileType: string; Width, Height, OX, OY: Integer);
      function GetImageName: string;
      procedure Origins(out Ox, Oy: Integer);
      function ImageIsOK: Boolean;

      property Image[Pos: Cardinal]: TFPImageBitmap read GetImage; default;
      property Count: Integer read GetImagesCount;
  end;

  { TGraphicLibrary }

  TGraphicLibraryCFLine = record
    Offset: Integer;
    Variable: string;
    Value: string;
  end;

  TGraphicLibraryItemConfig = record
    FileName: string;
    Name: string;
    Width: Integer;
    Height: Integer;
    OriginX: Integer;
    OriginY: Integer;
    FileType: string;
  end;

  TGraphicLibrary = class
    private
      Items: array of TGraphicItem;
      HashCorrectionStep: Cardinal;
      TableSize: Cardinal;
      LangLib: TApplicationTranslation;
      AppPath: string;

      function ScanLine(var TF: TextFile): TGraphicLibraryCFLine;
      procedure CreateHashTable(AwaitedLines: Integer);
      function AddToHashTable(ItemData: TGraphicLibraryItemConfig; Path: string): Boolean;
      function GetImageNamePosition(Name: string; Strict: Boolean=false): Cardinal;
      function GetGraphicItem(Name: string): TGraphicItem;
    public
      constructor Create(LL: TApplicationTranslation; Path: string);
      function LoadLibraryConfig(ConfigFile: string): Boolean;

      property Image[Name: string]: TGraphicItem read GetGraphicItem; default;
  end;

function RenderTextBitmap(Text: string; TextRed, TextGreen, TextBlue, BgRed, BgGreen, BgBlue: Byte; FH: Integer; FS: TFontStyles): TBitmap;

procedure RotateMatrix(Angle: Double; IX,IY: Integer; out OX, OY: Integer);

procedure RotateMatrixTrans(Angle: Double; IX,IY: Integer; out OX, OY: Integer);

function RotateImage(Image: TFPImageBitmap; Angle: Double; OrigX, OrigY: Integer; out NewOrigX, NewOrigY: Integer): TFPImageBitmap;

procedure DrawWithOrigin(Canvas: TCanvas; Image: TFPImageBitmap; X, Y, OrigX, OrigY: Integer);

procedure DrawGraphicImage(Canvas: TCanvas; GL: TGraphicLibrary; ImageName: String; ImageNumber: Integer; X, Y: Integer; Rotate: Double=0);

procedure DrawTextWithBorder(Text: string; Font: TFont; BgColor: TColor; BorderWidth: Integer; BorderColor: TColor; CenterX, CenterY: Integer; Canvas: TCanvas);

implementation

function RenderTextBitmap(Text: string; TextRed, TextGreen, TextBlue, BgRed, BgGreen, BgBlue: Byte; FH: Integer; FS: TFontStyles): TBitmap;
var
  Size: TSize;
begin
  Result := TBitmap.Create;
  Result.PixelFormat := pf32bit;
  Result.Canvas.Brush.Color := RGBToColor(BgRed, BgGreen, BgBlue);
  Result.Canvas.Font.Color := RGBToColor(TextRed, TextGreen, TextBlue);
  Result.Canvas.Font.Height := FH;
  Result.Canvas.Font.Style := FS;
  Result.Canvas.Font.Quality := fqCleartypeNatural;
  Size := Result.Canvas.TextExtent(Text);
  Result.Width := Size.cx + 3;
  Result.Height := Size.cy + 3;
  Result.Canvas.FillRect(0,0,Result.Width,Result.Height);
  Result.Canvas.TextOut(1,1,Text);
end;

procedure RotateMatrix(Angle: Double; IX,IY: Integer; out OX, OY: Integer);
begin
  OX := round(IX * cos(degtorad(Angle)) - IY * sin(degtorad(Angle)));
  OY := round(IX * sin(degtorad(Angle)) + IY * cos(degtorad(Angle)));
end;

procedure RotateMatrixTrans(Angle: Double; IX,IY: Integer; out OX, OY: Integer);
begin
  OX := round(IX * cos(degtorad(Angle)) + IY * sin(degtorad(Angle)));
  OY := round(- IX * sin(degtorad(Angle)) + IY * cos(degtorad(Angle)));
end;

function RotateImage(Image: TFPImageBitmap; Angle: Double; OrigX, OrigY: Integer; out NewOrigX, NewOrigY: Integer): TFPImageBitmap;
var
  NewWidth, NewHeight, X, Y, Nx, Ny: Integer;
  Pos, NewPos: record
    TopRX, TopRY, TopLX, TopLY: Integer;
    BottomRX, BottomRY, BottomLX, BottomLY: Integer;
  end;
  ResultIntf, InputImgIntf: TLazIntfImage;
  Line, LineOrig: PRGBAColorLine;
  LineOrig24Bit: PRGBColorLine;
begin
  Pos.TopLX := - OrigX;
  Pos.TopLY := OrigY;
  Pos.TopRX := Image.Width - OrigX;
  Pos.TopRY := OrigY;
  Pos.BottomLX := - OrigX;
  Pos.BottomLY := OrigY - Image.Height;
  Pos.BottomRX := Image.Width - OrigX;
  Pos.BottomRY := OrigY - Image.Height;

  RotateMatrixTrans(Angle, Pos.TopLX, Pos.TopLY, NewPos.TopLX, NewPos.TopLY);
  RotateMatrixTrans(Angle, Pos.TopRX, Pos.TopRY, NewPos.TopRX, NewPos.TopRY);
  RotateMatrixTrans(Angle, Pos.BottomLX, Pos.BottomLY, NewPos.BottomLX, NewPos.BottomLY);
  RotateMatrixTrans(Angle, Pos.BottomRX, Pos.BottomRY, NewPos.BottomRX, NewPos.BottomRY);

  NewHeight := Max(Max(NewPos.TopLY, NewPos.TopRY),Max(NewPos.BottomRY, NewPos.BottomLY)) - Min(Min(NewPos.TopLY, NewPos.TopRY),Min(NewPos.BottomRY, NewPos.BottomLY)) + 2;
  NewWidth := Max(Max(NewPos.TopLX, NewPos.TopRX),Max(NewPos.BottomRX, NewPos.BottomLX)) - Min(Min(NewPos.TopLX, NewPos.TopRX),Min(NewPos.BottomRX, NewPos.BottomLX)) + 2;

  NewOrigX := - Min(Min(NewPos.TopLX, NewPos.TopRX),Min(NewPos.BottomRX, NewPos.BottomLX)) + 1;
  NewOrigY := Max(Max(NewPos.TopLY, NewPos.TopRY),Max(NewPos.BottomRY, NewPos.BottomLY)) + 1;

  Result := TPortableNetworkGraphic.Create;
  Result.PixelFormat := pf32bit;
  Result.Width := NewWidth;
  Result.Height := NewHeight;

  Result.Canvas.Brush.Color := clWhite;
  Result.Canvas.FillRect(0, 0, NewWidth, NewHeight);

  ResultIntf := Result.CreateIntfImage;
  InputImgIntf := Image.CreateIntfImage;

  for Y := 0 to NewHeight - 1 do
  begin
    Line := ResultIntf.GetDataLineStart(Y);
    for X := 0 to NewWidth - 1 do
    begin
      Line^[X].Red := 0;
      Line^[X].Green := 0;
      Line^[X].Blue := 0;
      Line^[X].Alpha := 0;
      RotateMatrixTrans(Angle, X - NewOrigX, Y - NewOrigY, Nx, Ny);
      Nx := Nx + OrigX;
      Ny := Ny + OrigY;
      if (Nx >= 0) and (Nx <= Image.Width - 1) and (Ny >= 0) and (Ny <= Image.Height - 1) then
      begin
        if Image.PixelFormat = pf32bit then
        begin
          LineOrig := InputImgIntf.GetDataLineStart(Ny);
          Line^[X].Red := LineOrig^[Nx].Red;
          Line^[X].Green := LineOrig^[Nx].Green;
          Line^[X].Blue := LineOrig^[Nx].Blue;
          Line^[X].Alpha := LineOrig^[Nx].Alpha;
        end
        else
        begin
          LineOrig24Bit := InputImgIntf.GetDataLineStart(Ny);
          Line^[X].Red := LineOrig24Bit^[Nx].Red;
          Line^[X].Green := LineOrig24Bit^[Nx].Green;
          Line^[X].Blue := LineOrig24Bit^[Nx].Blue;
          Line^[X].Alpha := 255;
        end;
      end;
    end;
  end;

  Result.LoadFromIntfImage(ResultIntf);

  ResultIntf.Free;
  InputImgIntf.Free;
end;

procedure DrawWithOrigin(Canvas: TCanvas; Image: TFPImageBitmap; X, Y, OrigX, OrigY: Integer);
begin
  Canvas.Draw(X - OrigX, Y - OrigY, Image);
end;

procedure DrawGraphicImage(Canvas: TCanvas; GL: TGraphicLibrary; ImageName: String; ImageNumber: Integer; X, Y: Integer; Rotate: Double=0);
var
  Image: TFPImageBitmap;
  Ox, Oy: Integer;
begin
  if GL[ImageName] <> nil then
  begin
    Image := GL[ImageName][ImageNumber];
    if Image <> nil then
    begin
      GL[ImageName].Origins(Ox, Oy);
      if Rotate <> 0 then
      begin
        Image := RotateImage(Image, Rotate, Ox, Oy, Ox, Oy);
      end;
      DrawWithOrigin(Canvas, Image, X, Y, Ox, Oy);
      if (Rotate <> 0) and (Image <> GL[ImageName][ImageNumber]) then
        Image.Free;
    end;
  end;
end;

procedure DrawTextWithBorder(Text: string; Font: TFont; BgColor: TColor; BorderWidth: Integer; BorderColor: TColor; CenterX, CenterY: Integer; Canvas: TCanvas);
var
  TopX, TopY: Integer;
var
  BckBgColor, BckBorderColor: TColor;
  BckBrushStyle: TBrushStyle;
  BckPenWidth: Integer;
  BckFont: TFont;
  Size: TSize;
begin
  BckBgColor := Canvas.Brush.Color;
  BckBorderColor := Canvas.Pen.Color;
  BckPenWidth := Canvas.Pen.Width;
  BckBrushStyle := Canvas.Brush.Style;
  BckFont := Canvas.Font;

  Canvas.Font := Font;
  Canvas.Brush.Color := BgColor;
  Canvas.Brush.Style := bsSolid;
  Canvas.Pen.Color := BorderColor;
  Canvas.Pen.Width := BorderWidth;

  Size := Canvas.TextExtent(Text);
  TopX := round(CenterX - 2 - (BorderWidth div 2) - Size.cx / 2);
  TopY := round(CenterY - 2 - (BorderWidth div 2) - Size.cy / 2);

  Canvas.Rectangle(TopX, TopY, TopX + 4 + Size.cx + (BorderWidth div 2) * 2, TopY + 4 + Size.cy + (BorderWidth div 2) * 2);
  Canvas.TextOut(TopX + 2 + (BorderWidth div 2), TopY + 2 + (BorderWidth div 2), Text);

  Canvas.Font := BckFont;
  Canvas.Brush.Color := BckBgColor;
  Canvas.Brush.Style := BckBrushStyle;
  Canvas.Pen.Color := BckBorderColor;
  Canvas.Pen.Width := BckPenWidth;
end;

{ TGraphicItem }

function TGraphicItem.GetImage(Pos: Cardinal): TFPImageBitmap;
begin
  Result := nil;

  if Pos <= High(Images) then
    Result := Images[Pos];
end;

function TGraphicItem.GetImagesCount: Integer;
begin
  Result := Length(Images);
end;

constructor TGraphicItem.Create(FileName, Name, FileType: String; Width, Height, OX, OY: Integer);
var
  TempImg: TFPImageBitmap;
  I: Integer;
begin
  Images := nil;

  if not FileExists(FileName) then Exit;

  if FileType = 'png' then
    TempImg := TPortableNetworkGraphic.Create
  else if FileType = 'bmp' then
    TempImg := TBitmap.Create
  else
    Exit;

  OriginX := OX;
  OriginY := OY;
  ImageName := Name;

  try
    TempImg.LoadFromFile(FileName);
  except
    Exit;
  end;

  SetLength(Images, ceil(TempImg.Width / Width));

  if (Length(Images) = 1) then
  begin
    Images[0] := TempImg;
    Exit;
  end
  else
  begin
    for I := 0 to High(Images) do
    begin
      if FileType = 'png' then
        Images[I] := TPortableNetworkGraphic.Create
      else
        Images[I] := TBitmap.Create;

      Images[I].PixelFormat := TempImg.PixelFormat;
      Images[I].Width := Width;
      Images[I].Height := Height;
      Images[I].Canvas.Draw(0 - I * Width, 0, TempImg);
    end;
  end;

  TempImg.Free;
end;

function TGraphicItem.GetImageName: String;
begin
  Result := ImageName;
end;

procedure TGraphicItem.Origins(out Ox, Oy: Integer);
begin
  Ox := OriginX;
  Oy := OriginY;
end;

function TGraphicItem.ImageIsOK: Boolean;
var
  I: Integer;
begin
  Result := false;
  if Length(Images) = 0 then Exit;

  Result := true;

  for I := 0 to High(Images) do
    Result := Result and (Images[I] <> nil);
end;

{ TGraphicLibrary }

function TGraphicLibrary.ScanLine(var TF: TextFile): TGraphicLibraryCFLine;
var
  Line: String;
  I: Cardinal;
begin
  Result.Offset := 0;
  Result.Variable := '';
  Result.Value := '';

  ReadLn(TF, Line);

  for I := 1 to Length(Line) do
  begin
    if Line[I] <> ' ' then
      Break
    else
      Inc(Result.Offset);
  end;

  for I := I - 1 to Length(Line) do
  begin
    if Line[I] <> ':' then
      Result.Variable := Result.Variable + Line[I]
    else
      Break;
  end;

  Result.Variable := Trim(Result.Variable);
  Inc(I);

  Result.Value := Trim(Copy(Line, I, Length(Line)));
end;

procedure TGraphicLibrary.CreateHashTable(AwaitedLines: Integer);
var
  Exponent: SmallInt;
begin
  TableSize := Round(AwaitedLines * 2);

  for Exponent := 1 to 32 do
  begin
    if TableSize <= power(2,Exponent) then
      Break;
  end;

  TableSize := floor(power(2,Exponent));

  HashCorrectionStep := TableSize div 16;
  if HashCorrectionStep = 0 then
    HashCorrectionStep := 1;

  SetLength(Items, TableSize);
end;

function TGraphicLibrary.AddToHashTable(ItemData: TGraphicLibraryItemConfig; Path: String): Boolean;
var
  Pos: Cardinal;
begin
  Result := false;
  if ItemData.Name = '' then Exit;

  Pos := GetImageNamePosition(ItemData.Name);
  if Items[Pos] <> nil then Exit;
  if not ((ItemData.FileType = 'png') or (ItemData.FileType = 'bmp')) then Exit;

  if Path[Length(Path)] <> '/' then
    Path := Path + '/';

  Items[Pos] := TGraphicItem.Create(Path + ItemData.FileName, ItemData.Name, ItemData.FileType, ItemData.Width, ItemData.Height, ItemData.OriginX, ItemData.OriginY);
  Result := Items[Pos].ImageIsOK;
end;

function TGraphicLibrary.GetImageNamePosition(Name: string; Strict: Boolean=false): Cardinal;
var
  Sum, I, Cycles: Cardinal;
begin
  Result := 0;

  if TableSize = 0 then Exit;

  Sum := 0;

  for I := Length(Name) downto 1 do
  begin
    Sum := Sum + I * Ord(Name[I]);
  end;

  Result := Sum mod TableSize;
  if ((Items[Result] = nil) and (Strict = false)) or ((Items[Result] <> nil) and (Items[Result].GetImageName = Name)) then
    Exit;

  Cycles := 0;

  while not (((Items[Result] = nil) and (Strict = false)) or ((Items[Result] <> nil) and (Items[Result].GetImageName = Name))) do
  begin
    Result := (Result + HashCorrectionStep) mod TableSize;
    Inc(Cycles);
    if Cycles >= TableSize then
    begin
      Result := 0;
      Exit;
    end;
  end;
end;

function TGraphicLibrary.GetGraphicItem(Name: String): TGraphicItem;
var
  Pos: Cardinal;
begin
  Result := nil;

  if Length(Items) = 0 then Exit;

  Pos := GetImageNamePosition(Name, true);

  if Items[Pos] <> nil then
    Result := Items[Pos];
end;

constructor TGraphicLibrary.Create(LL: TApplicationTranslation; Path: string);
begin
  AppPath := Path;
  Items := nil;
  LangLib := LL;
end;

function TGraphicLibrary.LoadLibraryConfig(ConfigFile: string): Boolean;
var
  TF: TextFile;
  Line: TGraphicLibraryCFLine;
  Path: String;
  Section: String;
  ItemsInfo: array of TGraphicLibraryItemConfig;
  I: Integer;
  LineNumber: Cardinal;
begin
  Result := false;
  if not FileExists(AppPath + '\' + ConfigFile) then
  begin
    MessageDlg(LangLib['appformcaption'], StringReplace(LangLib['graphic_lib_file_to_found'], '#file#', ConfigFile, [rfReplaceAll, rfIgnoreCase]), mtError, [mbOK], 0);
    Exit;
  end;

  try
    Section := '';
    ItemsInfo := nil;

    Assign(TF, AppPath + '\' + ConfigFile);
    Reset(TF);

    LineNumber := 0;
    while not SeekEOF(TF) do
    begin
      Inc(LineNumber);
      Line := ScanLine(TF);
      if (Line.Offset = 0) and (Line.Variable = 'path') then
        Path := Line.Value
      else if (Line.Offset = 0) and (Line.Variable = 'items') then
        Section := 'items'
      else if ((Section = 'items') or (Section = 'items-item')) and (Line.Offset = 1) and (Line.Variable = 'item') then
      begin
        Section := 'items-item';
        SetLength(ItemsInfo, Length(ItemsInfo) + 1);
      end
      else if (Section = 'items-item') and (Line.Offset = 2) and (Line.Variable = 'name') then
        ItemsInfo[High(ItemsInfo)].Name := Line.Value
      else if (Section = 'items-item') and (Line.Offset = 2) and (Line.Variable = 'type') then
        ItemsInfo[High(ItemsInfo)].FileType := Line.Value
      else if (Section = 'items-item') and (Line.Offset = 2) and (Line.Variable = 'filename') then
        ItemsInfo[High(ItemsInfo)].FileName := Line.Value
      else if (Section = 'items-item') and (Line.Offset = 2) and (Line.Variable = 'width') then
        ItemsInfo[High(ItemsInfo)].Width := StrToIntDef(Line.Value, 0)
      else if (Section = 'items-item') and (Line.Offset = 2) and (Line.Variable = 'height') then
        ItemsInfo[High(ItemsInfo)].Height := StrToIntDef(Line.Value, 0)
      else if (Section = 'items-item') and (Line.Offset = 2) and (Line.Variable = 'originx') then
        ItemsInfo[High(ItemsInfo)].OriginX := StrToIntDef(Line.Value, 0)
      else if (Section = 'items-item') and (Line.Offset = 2) and (Line.Variable = 'originy') then
        ItemsInfo[High(ItemsInfo)].OriginY := StrToIntDef(Line.Value, 0)
      else
      begin
        MessageDlg(LangLib['appformcaption'], StringReplace(StringReplace(LangLib['graphic_lib_wrong_format'], '#line#', IntToStr(LineNumber), [rfReplaceAll, rfIgnoreCase]), '#file#', ConfigFile, [rfReplaceAll, rfIgnoreCase]), mtError, [mbOK], 0);
        Exit;
      end;

      //ShowMessage(IntToStr(Line.Offset) + ' ' + Line.Variable + ' ' + Line.Value);
    end;

    Close(TF);
    Result := true;
  except
    Close(TF);
    Result := false;
  end;

  CreateHashTable(Length(ItemsInfo) + 10);

  if Result then
  begin
    if Length(ItemsInfo) > 0 then
    begin
      for I := 0 to High(ItemsInfo) do
      begin
        Result := Result and AddToHashTable(ItemsInfo[I], AppPath + '\' + Path);
      end;
    end;
  end;
end;

end.

