unit machine;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics, ExtCtrls, Dialogs, appgraphic, math, syncobjs, translation, renderer;

type
  TStateEvent = class;

  TStateType = (mstNormal, mstAcceptState, mstAcceptAndEndState, mstFailAndEnd); // Typy stavov
  TStateEventType = (msetNormal, msetElse); // Typy udalosti
  TMachineAlphabet = set of Char; // Abeceda
  TEventArray = array of TStateEvent; // pole udalosti
  TAppOnAnimDoneEvent = procedure of object; // Udalost po skonceni animovania
  TAppSetOffsetEvent = procedure(Ox, Oy: Integer) of object; // Procedura na nastavenie offsetu zvonka

  { TState }

  TState = class
    private
      Events: TEventArray; // Pole udalosti
      StateType: TStateType; // Typ udalosti
      Ox, Oy: Integer; // Pozicia stavu na ploche
      AppImage: TImage; // Image objekt aplikacie
      Selected: Boolean; // Ci je stav vybrany alebo nie
      GraphicLibrary: TGraphicLibrary; // Kniznica grafiky
    public
      constructor Create(X, Y: Integer; Img: TImage; GL: TGraphicLibrary); // Konstruktor stavu
      procedure AppendEvent(Evnt: TStateEvent); // Pripne udalost ku stavu (pre nacitanie so suboru)
      procedure Render(OfsX, OfsY: Integer; Alphabet: TMachineAlphabet=[]); // Kreslenie stavu do grafickej plochy
      procedure MyXY(var X, Y: Integer); // Vrati cez parametre poziciu stavu v grafickej ploche
      procedure SetSelected(Sel: Boolean); // Nastavuje vlastnost vyberu stavu
      procedure SetXY(X, Y: Integer); // Nastavi novu poziciu stavu v grafickej ploche
      function NewEvent(ToState: TState):TStateEvent; // Vytvori novu udalost do daneho stavu
      function IsHereEventTowards(State: TState): Boolean; // Zisti, ci existuje udalost do daneho stavu
      procedure DeleteStateEvent(Event: TStateEvent); // Vymaze jednu konkretnu udalost
      procedure DeleteAllEvents; // Vymaze vsetky udalosti stavu
      function GetUsedSymbols(WithoutEvent: TStateEvent=nil): TMachineAlphabet; // Vrati vsetky symboli abecedy, ktore su priradene udalostiam vychadzajucim z tohoto stavu
      function CanHaveElseEvent(Alphabet: TMachineAlphabet): Boolean; // Vrati true, ak je este mozne pridat else event k tomuto stavu

      function Determine(Sentence: string): Boolean; // Vrati informaciu o akceptacii vety automatom v tomto alebo niektorom s nasledujucich stavov
      function DetermineEvent(Sentence: string): TStateEvent; // Vrati udalost, ktorou sa da pokracovat k dalsiemu stavu z tohoto vrchola

      property EventArray: TEventArray read Events; // Vrati kompletne vsetky udalosti tohoto stavu ako pole
      property TypeOfState: TStateType read StateType write StateType;
  end;

  { TStateEvent }

  TStateEvent = class
    private
      StateFrom: TState; // Informacia, z ktoreho stavu tato udalost vychadza
      StateTo: TState; // Informacia, do ktoreho stavu tato udalost smeruje
      EventType: TStateEventType; // Typ udalosti
      Symbols: TMachineAlphabet; // Symboly abecedy pre tuto udalost
      AppImage: TImage; // Image objekt aplikacie
      GraphicLibrary: TGraphicLibrary; // Graficka kniznica
      Angle: Double; // Uhol
      Radius: Integer; // Polomer kruznice spatnej udalosti
      ArrowCache: TPortableNetworkGraphic; // Cache pre zrotovany obrazok sipky smeru udalosti
      ArrowOriginX, ArrowOriginY: Integer; // Suradnice stredu sipky smeru udalosti
      Selected: Boolean; // Ci je alebo nie je tato udalost vybrana
      DisplayLabel: Boolean; // Ci sa ma alebo nema zobrazovat stitok s pouzitymi symbolmi

      function CalculateAngle: Double; // Prepocita uhol
      procedure CacheArrow; // Vytvori/obnovi cache pre sipku stavu
    public
      constructor Create(SFrom, STo: TState; Image: TImage; GL: TGraphicLibrary); // Konstruktor udalosti
      destructor Destroy; override; // Destruktor udalosti
      procedure Render(OfsX,OfsY: Integer); // Vykreslenie smerovej ciary udalosti
      procedure RenderLabel(OfsX,OfsY: Integer; FullAlphabet: TMachineAlphabet); // Vykresli stitok s pouzitymi symbolmi abecedy
      procedure RefreshCache; // Dava prikaz na celkove prepocitanie cache
      procedure SetAngle(NewAngle: Integer); // Nastavi zvonku zadany novy uhol (funguje len pre cyklicku udalost)
      procedure SetRadius(NewRadius: Integer); // Nastavi polomer kruznice pre cyklicku udalost

      function TestSelection(X, Y: Integer): Boolean; // Zistuje, ci bolo kliknute na tuto udalost
      procedure SetSelected(Status: Boolean); // Nastavuje vyber tejto udalosti

      function AnimationCanContinue(Count: Integer): Boolean; // Vrati info, ci animacia udalosti moze pokracovat, alebo sa ma ist animovat stav
      procedure RenderAnimatedSpark(OfsX, OfsY, Phase, Count: Integer); // Vykresli animovanu iskru podla fazy a pocitadla a povie, ci je este mozne pokracovat
      procedure AnimationSetOffset(Count: Integer; OffsetProc: TAppSetOffsetEvent); // Nastavi koordinaty offsetu pre animovanie

      property TypeOfEvent: TStateEventType read EventType write EventType;
      property UsedSymbols: TMachineAlphabet read Symbols write Symbols; // Vrati alebo zapise, ake symboly abecedy su pouzite v tejto udalosti
      property Target: TState read StateTo; // Vrati stav, do ktoreho udalost smeruje
      property Origin: TState read StateFrom; // Vrati stav, z ktoreho udalost vychadza
      property LabelEnabled: Boolean read DisplayLabel write DisplayLabel;
  end;

  { TMachine }

  TMachine = class
    private
      States: array of TState; // Pole vsetkych stavov
      StateEvents: TEventArray; // Pole vsetkych udalosti
      Alphabet: TMachineAlphabet; // Celkovy rozsah abecedy automatu
      AppImage: TImage; // Image objekt aplikacie
      SelState: TState; // Aktualne vybrany stav
      LastSelState: TState; // Minuly vybrany stav
      SelStateEvent: TStateEvent; // Prave vybrana udalost
      GraphicLibrary: TGraphicLibrary; // Graficka kniznica
      StartArrowAngle: Double;
      AppOnAnimDone: TAppOnAnimDoneEvent;
      AppEventSetOffset: TAppSetOffsetEvent;
      Animating: Boolean;
      CurrentAnimatedObject: TObject;
      Renderer: TRenderer;

      Animation: record
        State_Phase: Integer;
        State_Angle: Integer;
        State_Count: Integer;
        Event_Phase: Integer;
        Event_Count: Integer;
        Sentence: string;
      end;

      function TestCollision(X, Y: Integer; DontCheckState: TState=nil): Boolean; // Test, ci na suradniciach kliknutia dochadza ku kolizii s nejakym stavom
      function TestStateSelection(X, Y: Integer): TState; // Testuje vyber stavu
      function TestStateEventSelection(X, Y: Integer): TStateEvent; // Testuje vyber udalosti

      procedure MarkStateAsSelected(State: TState); // Oznaci dany stav ako vybrany, ostatne oznaci ako nevybrane
      procedure MarkStateEventAsSelected(StateEvent: TStateEvent); // Oznaci danu udalost ako vybranu, ostatne oznaci ako nevybrane
      procedure UnboundEvents(Events: array of TStateEvent); // Odpoji vsetky definovane udalosti z pola udalosti

      procedure AnimState_Start(var Render: Boolean); // Zaciatok animovania stavu
      procedure AnimState_Next(var Render: Boolean); // Dalsia cast animovania stavu
      procedure AnimEvent_Start(var Render: Boolean); // Zaciatok animovania udalosti
      procedure AnimEvent_Next(var Render: Boolean); // Pokracovanie animovania udalosti

      procedure AnimationEnd; // Koniec animacie
    public
      constructor Create(Image: TImage; Alb: TMachineAlphabet; GL: TGraphicLibrary); // Konstruktor automatu
      constructor Create(Image: TImage; GL: TGraphicLibrary); // Konstruktor bez zadania abecedy (ta sa bude nacitat so suboru
      procedure CreateState(X, Y: Integer; SnapToGrid: Boolean=false); // Vytvori novy stav
      procedure CreateStateEvent(Symbols: string; IsElse: Boolean=false); // Vytvori novu udalost
      function TestEventBetween: Boolean; // Vrati true, ak sa medzi dvoma vybranymi vrcholmi moze vytvorit nova udalost
      function TestElseEventPossibility: Boolean; // Vrati true, ak je mozne medzi dvoma stavmi vytvorit else udalost
      function TestMyStateHasElseEvent: Boolean; // Vrati true, ak vychodzi stav pre vybranu udalost ma nejaku udalost typu msetElse
      function GetAvailableSymbolsForEvent: TMachineAlphabet; // Vrati vsetky dostupne symboli pre prvy vybrany stav na spojenie s druhym
      procedure Render(OfsX, OfsY: Integer); // Vykresluje automat na graficku plochu
      function TestSelection(X, Y: Integer):Boolean; // Test vyberu stavov a udalosti
      procedure SetXYOfSelState(X, Y: Integer; SnapToGrid: Boolean=false); // Nastavi vybranemu stavu novu poziciu X a Y
      procedure DeleteSelected; // Vymaze vybrany objekt
      procedure SelectNone; // Zrusi vyber
      function TestIfMachineIsComplete: Boolean; // Vrati true, ak je automat dokonceny a moze sa pouzit
      function ReturnSelEventSymbols: string; // Vrati vsetky pouzite symboli vybranej udalosti ako retazec
      function IsStateSelected: Boolean; // Vrati true, ak je prave vybrany nejaky stav
      function IsFirstStateSelected: Boolean; // Vrati true, ak je vybrany zaciatocny stav
      function IsEventSelected: Boolean; // Vrati true, ak je prave oznacena udalost
      function IsCyclicEventSelected: Boolean; // Vrati true, ak je vybrana udalost smerujuca naspat do svojho vychodiskoveho stavu
      function IsElseEventSelected: Boolean; // Vrati true, ak je vybrana udalost typu msetElse
      function IsEventLabelDisplaing: Boolean; // Vrati true, ak sa zobrazuje stitok symbolov vybranej udalosti
      procedure SetFirstStateArrowAngle(Angle: Integer); // Nastavi uhol sipky prveho stavu
      procedure MakeStateFirst; // Nastavi vybrany stav ako vychodzi
      procedure SetSelStateType(NewType: TStateType); // Nastavi typ stavu, ktory je prave oznaceny
      function GetAvailableSymbolsForEdit: TMachineAlphabet; // Vrati zoznam dostupnych symbolov kvoli editacii vybranej udalosti
      procedure SetSelEventNewSymbols(Symbols: TMachineAlphabet); // Vybranej udalosti nastavi nove symboly
      procedure SetSelEventAngle(Angle: Integer); // Nastavi vybranej udalosti uhol (ak ide o cyklicku udalost)
      procedure SetSelEventRadius(Radius: Integer); // Nastavi vybranej udalosti polomer kruznice (ak ide o cyklicku udalost)
      procedure SetSelEventLabelDisplaing(Enabled: Boolean); // Nastavi vybranej udalosti zobrazovanie stitku symbolov

      function Determine(Sentence: string): Boolean; // Vyhodnoti vetu automatom a vrati true, ak je akceptovana
      procedure StartAnimation(Sentence: string; OnDone: TAppOnAnimDoneEvent; SetOffset: TAppSetOffsetEvent; Rnd: TRenderer); // Spusti animovanie automatu
      procedure ForceStopAnimation; // Vynuti zastavenie animacie

      function SaveToFile(FullFileName: string; Trans: TApplicationTranslation): Boolean; // Ulozi automat do suboru
      function LoadFromFile(FullFileName: string; Trans: TApplicationTranslation): Boolean; // Nacita automat so suboru

      procedure NumberOfStatesAndEvents(out St, Ev: Integer); // Cez parametre vrati pocet existujucich stavov a udalosti
      destructor Destroy; override; // Destruktor automatu

      property GetAlphabet: TMachineAlphabet read Alphabet;
      property RemainingSentence: string read Animation.Sentence;
  end;

var
  SyncInternalDataManipulation: TCriticalSection;

implementation

const FILE_CONTENT_STATES = $0001;
const FILE_CONTENT_EVENTS = $0002;
const FILE_CONTENT_MACHINE = $0004;

{ TState }

constructor TState.Create(X, Y: Integer; Img: TImage; GL: TGraphicLibrary);
begin
  Ox := X;
  Oy := Y;
  AppImage := Img;
  GraphicLibrary := GL;
end;

procedure TState.AppendEvent(Evnt: TStateEvent);
begin
  SetLength(Events, Length(Events) + 1);
  Events[High(Events)] := Evnt;
end;

procedure TState.Render(OfsX, OfsY: Integer; Alphabet: TMachineAlphabet=[]);
var
  ID: Integer;
begin
  ID := 0;
  if not ((CanHaveElseEvent(Alphabet) and (Alphabet - GetUsedSymbols = [])) or (not CanHaveElseEvent(Alphabet))) then
    ID := 2;
  if Selected then
    ID := 1;
  if StateType = mstNormal then
    DrawGraphicImage(AppImage.Canvas, GraphicLibrary, 'State_normal', ID, Ox + OfsX, Oy + OfsY)
  else if StateType = mstAcceptState then
    DrawGraphicImage(AppImage.Canvas, GraphicLibrary, 'State_accept', ID, Ox + OfsX, Oy + OfsY)
  else if StateType = mstAcceptAndEndState then
    DrawGraphicImage(AppImage.Canvas, GraphicLibrary, 'State_accept_and_end', ID, Ox + OfsX, Oy + OfsY)
  else if StateType = mstFailAndEnd then
    DrawGraphicImage(AppImage.Canvas, GraphicLibrary, 'State_fail_and_end', ID, Ox + OfsX, Oy + OfsY);
end;

procedure TState.MyXY(var X, Y: Integer);
begin
  X := Ox;
  Y := Oy;
end;

procedure TState.SetSelected(Sel: Boolean);
begin
  Selected := Sel;
end;

procedure TState.SetXY(X, Y: Integer);
begin
  Ox := X;
  Oy := Y;
end;

function TState.NewEvent(ToState: TState):TStateEvent;
begin
  SyncInternalDataManipulation.Enter; // Vstup do kritickej sekcie
  try
    SetLength(Events, Length(Events) + 1); // Rozsiri interne pole udalosti
    Events[High(Events)] := TStateEvent.Create(Self, ToState, AppImage, GraphicLibrary); // Vytvori novu udalost
    Result := Events[High(Events)]; // Vrati referenciu na novo vytvorenu udalost
  finally
    SyncInternalDataManipulation.Leave; // Vystup z kritickej sekcie
  end;
end;

function TState.IsHereEventTowards(State: TState): Boolean;
var
  I: Integer;
begin
  SyncInternalDataManipulation.Enter; // Vstup do kritickej sekcie
  Result := false; // Nic sa este nenaslo
  try
    if Length(Events) > 0 then // Ak tu su nejake udalosti
      for I := 0 to High(Events) do
        if Events[I].Target = State then // A nejaka udalost sa rovna tej hladanej
        begin
          Result := True; // Tak sme ju nasli
          Break; // A vyskocime z cyklu
        end;
  finally
    SyncInternalDataManipulation.Leave; // Vystup z kritickej sekcie
  end;
end;

procedure TState.DeleteStateEvent(Event: TStateEvent);
var
  I: integer;
begin
  SyncInternalDataManipulation.Enter; // Vstup do kritickej sekcie
  try
    if Length(Events) > 0 then // Ak tu nejake udalosti su
      for I := 0 to High(Events) do
      begin
        if Events[I] = Event then // A nasli sme tu, ktoru chceme zmazat
        begin
          FreeAndNil(Events[I]); // Uvolnime ju z pamate
          Events[I] := Events[High(Events)]; // Na jej miesto dame poslednu s pola udalosti
          SetLength(Events, Length(Events) - 1); // Skratime dlzku pola o jedna
          Break; // A vyskocime s cyklu
        end;
      end;
  finally
    SyncInternalDataManipulation.Leave; // Vystup z kritickej sekcie
  end;
end;

procedure TState.DeleteAllEvents;
var
  I: integer;
begin
  SyncInternalDataManipulation.Enter; // Vstup do kritickej sekcie
  try
    if Length(Events) > 0 then // Ak udalosti v tomto stave existuju
      for I := 0 to High(Events) do
      begin
        FreeAndNil(Events[I]); // Vsetky od prvej po poslednu uvolnime z pamate
      end;
    SetLength(Events, 0); // A pole udalosti skratime na nulu
  finally
    SyncInternalDataManipulation.Leave; // Vystup z kritickej sekcie
  end;
end;

function TState.GetUsedSymbols(WithoutEvent: TStateEvent=nil): TMachineAlphabet;
var
  I: Integer;
begin
  Result := []; // Nic nie je priradene
  SyncInternalDataManipulation.Enter; // Vstup do kritickej sekcie
  try
    if Length(Events) > 0 then // Ak nejake udalosti existuju
      for I := 0 to High(Events) do // Vsetky prejdi
        if (Events[I].TypeOfEvent <> msetElse) and (Events[I] <> WithoutEvent) then // Pokial tato udalost nie je typu Else alebo nie je oznazena za vynatu
          Result := Result + Events[I].UsedSymbols; // A spocitaj mnoziny pouzitych znakov
  finally
    SyncInternalDataManipulation.Leave; // Vystup z kritickej sekcie
  end;
end;

function TState.CanHaveElseEvent(Alphabet: TMachineAlphabet): Boolean;
var
  I: Integer;
begin
  Result := false;
  if Alphabet - GetUsedSymbols = [] then Exit;

  SyncInternalDataManipulation.Enter;
  try
    if Length(Events) > 0 then
      for I := 0 to High(Events) do
        if Events[I].TypeOfEvent = msetElse then
        begin
          SyncInternalDataManipulation.Leave;
          Exit;
        end;
  finally
    SyncInternalDataManipulation.Leave;
  end;

  Result := true;
end;

function TState.Determine(Sentence: string): Boolean;
var
  NextEvent: TStateEvent;
begin
  if Sentence = '' then
  begin
    if (StateType = mstNormal) or (StateType = mstFailAndEnd) then
    begin
      Result := false;
      Exit;
    end
    else
    begin
      Result := true;
      Exit;
    end;
  end
  else
  begin
    if StateType = mstAcceptAndEndState then
    begin
      Result := true;
      Exit;
    end
    else if StateType = mstFailAndEnd then
    begin
      Result := false;
      Exit;
    end
    else
    begin
      NextEvent := DetermineEvent(Sentence);
      if NextEvent = nil then
      begin
        Result := false;
        Exit;
      end;

      Result := NextEvent.Target.Determine(Copy(Sentence, 2, Length(Sentence) - 1));
    end;
  end;
end;

function TState.DetermineEvent(Sentence: string): TStateEvent;
var
  TempElseEvent: TStateEvent;
  I: Integer;
begin
  TempElseEvent := nil;
  Result := nil;
  if Sentence = '' then Exit;
  if (StateType = mstFailAndEnd) or (StateType = mstAcceptAndEndState) then Exit;

  if Length(Events) > 0 then
    for I := 0 to High(Events) do
      if Events[I].TypeOfEvent = msetElse then
        TempElseEvent := Events[I]
      else if Sentence[1] in Events[I].UsedSymbols then
      begin
        Result := Events[I];
        Exit;
      end;

  Result := TempElseEvent;
end;

{ TStateEvent }

function TStateEvent.CalculateAngle: Double;
var
  Fx, Fy, Tx, Ty: Integer;
begin
  if StateFrom = StateTo then
  begin
    Result := Angle;
    Exit;
  end;

  Fx := 0; Fy := 0; Tx := 0; Ty := 0;
  Result := 0;
  StateFrom.MyXY(Fx, Fy);
  StateTo.MyXY(Tx, Ty);

  Tx := Tx - FX;
  Ty := FY - Ty;
  if Ty = 0 then
    if Tx = 0 then
      Result := 0
    else if Tx < 0 then
      Result := 270
    else
      Result := 90
  else if Ty > 0 then
    if Tx >= 0 then
      Result := radtodeg(ArcTan(Tx/Ty))
    else
      Result := 360 - radtodeg(ArcTan(-Tx/Ty))
  else if Tx >= 0 then
    Result := 180 - radtodeg(ArcTan(-Tx/Ty))
  else
    Result := 180 + radtodeg(ArcTan(Tx/Ty));
end;

procedure TStateEvent.CacheArrow;
var
  ID: Byte;
begin
  if ArrowCache <> nil then
    ArrowCache.Free;

  ID := 0;
  if Selected then
    ID := 1;

  if (GraphicLibrary['StateEventLineArrow'] <> nil) and (GraphicLibrary['StateEventLineArrow'][ID] <> nil) then
  begin
    GraphicLibrary['StateEventLineArrow'].Origins(ArrowOriginX, ArrowOriginY);
    ArrowCache := TPortableNetworkGraphic(RotateImage(GraphicLibrary['StateEventLineArrow'][ID], Angle, ArrowOriginX, ArrowOriginY, ArrowOriginX, ArrowOriginY));
  end;
end;

constructor TStateEvent.Create(SFrom, STo: TState; Image: TImage;
  GL: TGraphicLibrary);
begin
  Radius := 30;
  Angle := 0;
  StateFrom := SFrom;
  StateTo := STo;
  AppImage := Image;
  GraphicLibrary := GL;
  DisplayLabel := true;
  RefreshCache;
end;

destructor TStateEvent.Destroy;
begin
  Symbols := [];
  if ArrowCache <> nil then
    FreeAndNil(ArrowCache);
  inherited Destroy;
end;

procedure TStateEvent.Render(OfsX, OfsY: Integer);
var
  Fx,Fy,Tx,Ty,Cx,Cy: Integer;
  Font: TFont;
begin
  Fx := 0; Fy := 0; Tx := 0; Ty := 0;

  Font := TFont.Create;

  Font.Color := clBlue;
  if Selected then
    Font.Color := clRed;
  Font.Height := 16;
  Font.Name := 'Courier New';

  if StateFrom <> StateTo then
  begin
    StateFrom.MyXY(Fx, Fy);
    StateTo.MyXY(Tx, Ty);
    if StateTo.IsHereEventTowards(StateFrom) then
    begin
      RotateMatrix(Angle, 10, 0, Cx, Cy);
    end
    else
    begin
      Cx := 0;
      Cy := 0;
    end;

    AppImage.Canvas.Pen.Color := clRed;
    if Selected then
       AppImage.Canvas.Pen.Color := clGreen;
    AppImage.Canvas.Pen.Width := 1;
    if Selected then
      AppImage.Canvas.Pen.Width := 3;
    AppImage.Canvas.Pen.Style := psSolid;
    if EventType = msetElse then
      AppImage.Canvas.Pen.Style := psDashDot;
    AppImage.Canvas.MoveTo(Fx + OfsX + Cx, Fy + OfsY + Cy);
    AppImage.Canvas.LineTo(Tx + OfsX + Cx, Ty + OfsY + Cy);

    DrawWithOrigin(AppImage.Canvas, ArrowCache, Fx - (Fx - Tx) div 2 + OfsX + Cx, Fy - (Fy - Ty) div 2 + OfsY + Cy, ArrowOriginX, ArrowOriginY);
    DrawWithOrigin(AppImage.Canvas, ArrowCache, Fx - (Fx - Tx) div 4 + OfsX + Cx, Fy - (Fy - Ty) div 4 + OfsY + Cy, ArrowOriginX, ArrowOriginY);
    DrawWithOrigin(AppImage.Canvas, ArrowCache, Fx - round((Fx - Tx) * 0.75) + OfsX + Cx, Fy - round((Fy - Ty) * 0.75) + OfsY + Cy, ArrowOriginX, ArrowOriginY);
  end
  else
  begin
    StateFrom.MyXY(Fx, Fy);
    Cx := 0;
    Cy := - Radius;
    RotateMatrix(Angle - 45, Cx, Cy, Cx, Cy);

    AppImage.Canvas.Pen.Color := clRed;
    if Selected then
       AppImage.Canvas.Pen.Color := clGreen;
    AppImage.Canvas.Pen.Width := 1;
    if Selected then
      AppImage.Canvas.Pen.Width := 3;
    AppImage.Canvas.Pen.Style := psSolid;
    if EventType = msetElse then
      AppImage.Canvas.Pen.Style := psDashDot;
    AppImage.Canvas.Brush.Style := bsClear;
    AppImage.Canvas.Ellipse(Cx + OfsX + Fx - Radius, Cy + OfsY + Fy - Radius, Cx + OfsX + Fx + Radius, Cy + OfsY + Fy + Radius);
    AppImage.Canvas.Brush.Style := bsSolid;
  end;

  AppImage.Canvas.Pen.Style := psSolid;

  Font.Free;
end;

procedure TStateEvent.RenderLabel(OfsX, OfsY: Integer; FullAlphabet: TMachineAlphabet);
var
  Fx,Fy,Tx,Ty,Cx,Cy,Lx,Ly: Integer;
  Font: TFont;
  I: Char;
  Symbs: string;
  TempAlpha: TMachineAlphabet;
  E: Integer;
begin
  if not DisplayLabel then Exit;

  Fx := 0; Fy := 0; Tx := 0; Ty := 0;

  Symbs := '';
  if EventType = msetNormal then
  begin
    for I := chr(0) to chr(255) do
      if I in Symbols then
        Symbs := Symbs + I;
  end
  else
  begin
    TempAlpha := [];
    for E := 0 to High(Origin.EventArray) do
      TempAlpha := TempAlpha + Origin.EventArray[E].UsedSymbols;
    TempAlpha := TempAlpha - UsedSymbols;

    TempAlpha := FullAlphabet - TempAlpha;
    for I := chr(0) to chr(255) do
      if I in TempAlpha then
        Symbs := Symbs + I;
  end;

  Font := TFont.Create;

  Font.Color := clBlue;
  if Selected then
    Font.Color := clRed;
  Font.Height := 16;
  Font.Name := 'Courier New';

  if StateFrom <> StateTo then
  begin
    StateFrom.MyXY(Fx, Fy);
    StateTo.MyXY(Tx, Ty);
    if StateTo.IsHereEventTowards(StateFrom) then
    begin
      RotateMatrix(Angle, 10, 0, Cx, Cy);
    end
    else
    begin
      Cx := 0;
      Cy := 0;
    end;
    Lx := 0; Ly := round(sqrt(sqr(Fx - Tx) + sqr(Fy - Ty)) * 0.33);
    if Ly < 30 then Ly := 30;
    RotateMatrixTrans(Angle, Lx, Ly, Lx, Ly);

    if EventType = msetNormal then
    begin
      if Selected then
        DrawTextWithBorder(Symbs, Font, clWhite, 3, clGreen, Fx + Lx + OfsX + Cx, Fy - Ly + OfsY + Cy, AppImage.Canvas)
      else
        DrawTextWithBorder(Symbs, Font, clWhite, 1, clRed, Fx + Lx + OfsX + Cx, Fy - Ly + OfsY + Cy, AppImage.Canvas);
    end
    else
    begin
      Font.Color := clRed;
      if Selected then
        Font.Color := clBlue;

      if Selected then
        DrawTextWithBorder(Symbs, Font, clWhite, 3, clGreen, Fx + Lx + OfsX + Cx, Fy - Ly + OfsY + Cy, AppImage.Canvas)
      else
        DrawTextWithBorder(Symbs, Font, clWhite, 1, clBlue, Fx + Lx + OfsX + Cx, Fy - Ly + OfsY + Cy, AppImage.Canvas);
    end;
  end
  else
  begin
    StateFrom.MyXY(Fx, Fy);
    Cx := 0;
    Cy := - Radius;
    RotateMatrix(Angle - 45, Cx, Cy, Cx, Cy);

    Lx := 0;
    Ly := -Radius;

    RotateMatrix(Angle, Lx, Ly, Lx, Ly);

    if EventType = msetNormal then
    begin
      if Selected then
        DrawTextWithBorder(Symbs, Font, clWhite, 3, clGreen, Cx + Fx + Lx + OfsX, Cy + Fy + Ly + OfsY, AppImage.Canvas)
      else
        DrawTextWithBorder(Symbs, Font, clWhite, 1, clRed, Cx + Fx + Lx + OfsX, Cy + Fy + Ly + OfsY, AppImage.Canvas);
    end
    else
    begin
      Font.Color := clRed;
      if Selected then
        Font.Color := clBlue;

      if Selected then
        DrawTextWithBorder(Symbs, Font, clWhite, 3, clGreen, Cx + Fx + Lx + OfsX, Cy + Fy + Ly + OfsY, AppImage.Canvas)
      else
        DrawTextWithBorder(Symbs, Font, clWhite, 1, clBlue, Cx + Fx + Lx + OfsX, Cy + Fy + Ly + OfsY, AppImage.Canvas);
    end;
  end;

  AppImage.Canvas.Pen.Style := psSolid;

  Font.Free;
end;

procedure TStateEvent.RefreshCache;
begin
  Angle := CalculateAngle;
  CacheArrow;
end;

procedure TStateEvent.SetAngle(NewAngle: Integer);
begin
  if Origin = Target then
    Angle := NewAngle;
end;

procedure TStateEvent.SetRadius(NewRadius: Integer);
begin
  if Origin = Target then
    Radius := NewRadius;
end;

function TStateEvent.TestSelection(X, Y: Integer): Boolean;
var
  Fx,Fy,Tx,Ty,Cx,Cy,Nx,Ny,Vx,Vy,Ix,Iy: Integer;
  T: Double;
begin
  Result := false;
  Fx := 0; Fy := 0; Tx := 0; Ty := 0;
  if StateFrom <> StateTo then
  begin
    StateFrom.MyXY(Fx, Fy);
    StateTo.MyXY(Tx, Ty);
    if StateTo.IsHereEventTowards(StateFrom) then
    begin
      RotateMatrix(Angle, 10, 0, Cx, Cy);
    end
    else
    begin
      Cx := 0;
      Cy := 0;
    end;

    Fx := Fx + Cx; Fy := Fy + Cy;
    Tx := Tx + Cx; Ty := Ty + Cy;

    Ny := - (Fx - Tx);
    Nx := Fy - Ty;
    Vx := Fx - Tx;
    Vy := Fy - Ty;

    T := (Vx * (Y - Fy) - Vy * (X - Fx)) / (Vy * Nx - Vx * Ny);
    Ix := round(X + Nx * T);
    Iy := round(Y + Ny * T);

    T := sqrt(sqr(Fx - Tx) + sqr(Fy - Ty)) / 2;
    Cx := Fx - (Fx - Tx) div 2;
    Cy := Fy - (Fy - Ty) div 2;

    if sqrt(sqr(X - Cx) + sqr(Y - Cy)) > T then Exit;

    if sqrt(sqr(X - Ix) + sqr(Y - Iy)) <= 5 then
      Result := true;
  end
  else
  begin
    StateFrom.MyXY(Fx, Fy);
    Cx := 0;
    Cy := - Radius;
    RotateMatrix(Angle - 45, Cx, Cy, Cx, Cy);

    if abs(round(sqrt(sqr(Cx + Fx - X) + sqr(Cy + Fy - Y))) - round(sqrt(sqr(Cx) + sqr(Cy)))) <= 5 then
      Result := true;
  end;
end;

procedure TStateEvent.SetSelected(Status: Boolean);
begin
  if Selected <> Status then
  begin
    Selected := Status;
    RefreshCache;
  end;
end;

function TStateEvent.AnimationCanContinue(Count: Integer): Boolean;
var
  OrX, OrY, TgX, TgY, Sx, Sy: Integer;
  LLength, SLenght: Double;
begin
  Result := true;
  if Origin = Target then
  begin
    if Count = 12 then
      Result := false;
  end
  else
  begin
    OrX := 0; OrY := 0; TgX := 0; TgY := 0;
    Origin.MyXY(OrX, OrY);
    Target.MyXY(TgX, TgY);
    Sx := 0; Sy := - 20 * Count;
    RotateMatrix(Angle, Sx, Sy, Sx, Sy);
    LLength := sqrt(sqr(OrX - TgX) + sqr(OrY - TgY));
    SLenght := sqrt(sqr(OrX - (OrX + Sx)) + sqr(OrY - (OrY + Sy)));
    if SLenght > LLength then
    begin
      Result := false;
    end
  end;
end;

procedure TStateEvent.RenderAnimatedSpark(OfsX, OfsY, Phase, Count: Integer);
var
  OrX, OrY, TgX, TgY, Sx, Sy, Cx, Cy: Integer;
  LLength, SLenght: Double;
begin
  if Origin = Target then
  begin
    OrX := 0; OrY := 0;
    Origin.MyXY(OrX, OrY);
    Cx := 0;
    Cy := - Radius;
    RotateMatrix(Angle - 45, Cx, Cy, Cx, Cy);
    LLength := Angle + 135 - 30 * Count;
    Sx := 0; Sy := - Radius;
    RotateMatrix(LLength, Sx, Sy, Sx, Sy);

    DrawGraphicImage(AppImage.Canvas, GraphicLibrary, 'Anim_event', Phase, OrX + Cx + OfsX + Sx, OrY + Cy + OfsY + Sy, LLength - 90);
  end
  else
  begin
    OrX := 0; OrY := 0; TgX := 0; TgY := 0;
    Origin.MyXY(OrX, OrY);
    Target.MyXY(TgX, TgY);
    if Target.IsHereEventTowards(Origin) then
    begin
      RotateMatrix(Angle, 10, 0, Cx, Cy);
    end
    else
    begin
      Cx := 0;
      Cy := 0;
    end;
    Sx := 0; Sy := - 20 * Count;
    RotateMatrix(Angle, Sx, Sy, Sx, Sy);
    LLength := sqrt(sqr(OrX - TgX) + sqr(OrY - TgY));
    SLenght := sqrt(sqr(OrX - (OrX + Sx)) + sqr(OrY - (OrY + Sy)));
    if SLenght > LLength then
    begin
      Sx := TgX;
      Sy := TgY;
    end
    else
    begin
      Sx := Sx + OrX;
      Sy := Sy + OrY;
    end;
    DrawGraphicImage(AppImage.Canvas, GraphicLibrary, 'Anim_event', Phase, Sx + OfsX + Cx, Sy + OfsY + Cy, Angle);
  end;
end;

procedure TStateEvent.AnimationSetOffset(Count: Integer;
  OffsetProc: TAppSetOffsetEvent);
var
  OrX, OrY, TgX, TgY, Sx, Sy, Cx, Cy: Integer;
  LLength, SLenght: Double;
begin
  if Origin = Target then
  begin
    OrX := 0; OrY := 0;
    Origin.MyXY(OrX, OrY);
    Cx := 0;
    Cy := - Radius;
    RotateMatrix(Angle - 45, Cx, Cy, Cx, Cy);
    LLength := Angle + 135 - 30 * Count;
    Sx := 0; Sy := - Radius;
    RotateMatrix(LLength, Sx, Sy, Sx, Sy);

    if Assigned(OffsetProc) then
      OffsetProc(OrX + Cx + Sx, OrY + Cy + Sy);
  end
  else
  begin
    OrX := 0; OrY := 0; TgX := 0; TgY := 0;
    Origin.MyXY(OrX, OrY);
    Target.MyXY(TgX, TgY);
    Sx := 0; Sy := - 20 * Count;
    RotateMatrix(Angle, Sx, Sy, Sx, Sy);
    LLength := sqrt(sqr(OrX - TgX) + sqr(OrY - TgY));
    SLenght := sqrt(sqr(OrX - (OrX + Sx)) + sqr(OrY - (OrY + Sy)));
    if SLenght > LLength then
    begin
      Sx := TgX;
      Sy := TgY;
    end
    else
    begin
      Sx := Sx + OrX;
      Sy := Sy + OrY;
    end;;
    if Assigned(OffsetProc) then
      OffsetProc(Sx, Sy);
  end;
end;

{ TMachine }

function TMachine.TestCollision(X, Y: Integer; DontCheckState: TState=nil): Boolean;
var
  I, Sx, Sy: Integer;
begin
  SyncInternalDataManipulation.Enter;
  Sx := 0; Sy := 0;
  Result := false;
  try
    if Length(States) > 0 then
      for I := 0 to High(States) do
      begin
        if States[I] = DontCheckState then Continue;
        States[I].MyXY(Sx,Sy);
        if sqrt(sqr(Sx - X) + sqr(Sy - Y)) < 100 then
        begin
          Result := true;
          Break;
        end;
      end;
  finally
    SyncInternalDataManipulation.Leave;
  end;
end;

function TMachine.TestStateSelection(X, Y: Integer): TState;
var
  I, Sx, Sy: Integer;
begin
  Sx := 0; Sy := 0;
  Result := nil;
  if Length(States) > 0 then
    for I := 0 to High(States) do
    begin
      States[I].MyXY(Sx, Sy);
      if sqrt(sqr(Sx - X) + sqr(Sy - Y)) < 15 then
      begin
        Result := States[I];
        Exit;
      end;
    end;
end;

function TMachine.TestStateEventSelection(X, Y: Integer): TStateEvent;
var
  I: Integer;
begin
  Result := nil;
  SyncInternalDataManipulation.Enter;
  try
    if Length(StateEvents) > 0 then
      for I := 0 to High(StateEvents) do
        if StateEvents[I].TestSelection(X, Y) then
        begin
          Result := StateEvents[I];
          Break;
        end;
  finally
    SyncInternalDataManipulation.Leave;
  end;
end;

procedure TMachine.MarkStateAsSelected(State: TState);
var
  I: Integer;
begin
  if Length(States) > 0 then
    for I := 0 to High(States) do
    begin
      if States[I] = State then
        States[I].SetSelected(true)
      else
        States[I].SetSelected(false);
    end;
end;

procedure TMachine.MarkStateEventAsSelected(StateEvent: TStateEvent);
var
  I: Integer;
begin
  if Length(StateEvents) > 0 then
    for I := 0 to High(StateEvents) do
    begin
      if StateEvents[I] = StateEvent then
        StateEvents[I].SetSelected(true)
      else
        StateEvents[I].SetSelected(false);
    end;
end;

procedure TMachine.SelectNone;
begin
  LastSelState := SelState;
  SelState := nil;
  MarkStateAsSelected(nil);
end;

function TMachine.TestIfMachineIsComplete: Boolean;
var
  I, E: Integer;
  Events: TEventArray;
  ElseEvent: Boolean;
  SetOfSymbols: TMachineAlphabet;
begin
  Result := false;
  if Length(States) = 0 then Exit;

  SyncInternalDataManipulation.Enter;
  try
    for I := 0 to High(States) do
    begin
      Events := States[I].EventArray;
      if Length(Events) = 0 then
      begin
        SyncInternalDataManipulation.Leave;
        Exit;
      end;
      SetOfSymbols := [];
      ElseEvent := false;
      for E := 0 to High(Events) do
      begin
        if Events[E].TypeOfEvent = msetElse then
          ElseEvent := true
        else
          SetOfSymbols := SetOfSymbols + Events[E].UsedSymbols;
      end;
      if not ((not ElseEvent and (Alphabet - SetOfSymbols = [])) or ElseEvent)  then
      begin
        SyncInternalDataManipulation.Leave;
        Exit;
      end;
    end;
  finally
    SyncInternalDataManipulation.Leave;
  end;

  Result := true;
end;

function TMachine.ReturnSelEventSymbols: string;
var
  I: Char;
begin
  Result := '';
  if SelStateEvent <> nil then
  begin
    for I := #0 to #255 do
      if I in SelStateEvent.UsedSymbols then
        Result := Result + I;
  end;
end;

function TMachine.IsStateSelected: Boolean;
begin
  Result := False;

  if SelState = nil then Exit;

  Result := true;
end;

function TMachine.IsFirstStateSelected: Boolean;
begin
  Result := False;

  if SelState = nil then Exit;

  if (Length(States) > 0) and (States[0] = SelState) then
    Result := true;
end;

function TMachine.IsEventSelected: Boolean;
begin
  Result := false;
  if SelStateEvent= nil then Exit;
  Result := true;
end;

function TMachine.IsCyclicEventSelected: Boolean;
begin
  Result := IsEventSelected;

  if Result then
    if SelStateEvent.Origin = SelStateEvent.Target then
      Result := true
    else
      Result := false;
end;

function TMachine.IsElseEventSelected: Boolean;
begin
  Result := IsEventSelected;

  if Result then
    if SelStateEvent.TypeOfEvent = msetElse then
      Result := true
    else
      Result := false;
end;

function TMachine.IsEventLabelDisplaing: Boolean;
begin
  Result := False;
  if SelStateEvent = nil then Exit;

  Result := SelStateEvent.LabelEnabled;
end;

procedure TMachine.SetFirstStateArrowAngle(Angle: Integer);
begin
  StartArrowAngle := Angle;
end;

procedure TMachine.MakeStateFirst;
var
  I: Integer;
begin
  if SelState = nil then Exit;

  SyncInternalDataManipulation.Enter;
  try
    if Length(States) > 0 then
      for I := 0 to High(States) do
        if States[I] = SelState then
        begin
          States[I] := States[0];
          States[0] := SelState;
          Break;
        end;
  finally
    SyncInternalDataManipulation.Leave;
  end;
end;

procedure TMachine.SetSelStateType(NewType: TStateType);
begin
  if SelState = nil then Exit;

  SelState.TypeOfState := NewType;
end;

function TMachine.GetAvailableSymbolsForEdit: TMachineAlphabet;
begin
  Result := [];

  if SelStateEvent = nil then Exit;

  Result := Alphabet - SelStateEvent.Origin.GetUsedSymbols(SelStateEvent);
end;

procedure TMachine.SetSelEventNewSymbols(Symbols: TMachineAlphabet);
begin
  if SelStateEvent = nil then Exit;

  SelStateEvent.UsedSymbols := Symbols;
end;

procedure TMachine.SetSelEventAngle(Angle: Integer);
begin
  if SelStateEvent = nil then Exit;

  SelStateEvent.SetAngle(Angle);
end;

procedure TMachine.SetSelEventRadius(Radius: Integer);
begin
  if SelStateEvent = nil then Exit;

  SelStateEvent.SetRadius(Radius);
end;

procedure TMachine.SetSelEventLabelDisplaing(Enabled: Boolean);
begin
  if SelStateEvent = nil then Exit;

  SelStateEvent.LabelEnabled := Enabled;
end;

function TMachine.Determine(Sentence: string): Boolean;
begin
  Result := false;
  if not TestIfMachineIsComplete then Exit;

  SelectNone;

  if Length(States) > 0 then
  begin
    SyncInternalDataManipulation.Enter;
    try
      Result := States[0].Determine(Sentence);
    finally
      SyncInternalDataManipulation.Leave;
    end;
  end;
end;

procedure TMachine.StartAnimation(Sentence: string; OnDone: TAppOnAnimDoneEvent; SetOffset: TAppSetOffsetEvent; Rnd: TRenderer);
var
  FireTime: TTimeStamp;
begin
  if Animating or (Length(States) = 0) then
  begin
    if Assigned(OnDone) then
      OnDone;
    Exit;
  end;
  Animating := True;
  CurrentAnimatedObject := States[0];
  Renderer := Rnd;
  Animation.Sentence := Sentence;
  AppEventSetOffset := SetOffset;
  AppOnAnimDone := OnDone;

  FireTime := TimeStampPlusMicrotime(DateTimeToTimeStamp(Now), 10);
  Renderer.AddAction(Self, @AnimState_Start, FireTime);
end;

procedure TMachine.ForceStopAnimation;
begin
  if Animating then
    Renderer.Sync(@AnimationEnd);
    //AnimationEnd;
end;

function TMachine.SaveToFile(FullFileName: string; Trans: TApplicationTranslation): Boolean;
var
  F: File;
  Stream: TStream;
  Str: String;
  Wrd: Word;
  Int, I, X, Y: Integer;
  Bool: Boolean;
  Bt: Byte;
  E: Char;
  Db: Double;

  function GetStateIndex(State: TState): Integer;
  begin
    if Length(States) > 0 then
      for Result := 0 to High(States) do
        if States[Result] = State then Exit;
    Result := -1;
  end;

begin
  Result := false;

  try
    AssignFile(F, FullFileName);
    Rewrite(F);
    CloseFile(F);
  except
    MessageDlg(Trans['machine_save_to_file_caption'], Trans['machine_save_to_file_error_unwritable'], mtError, [mbOK], 0);
    CloseFile(F);
    Exit;
  end;

  X := 0; Y := 0;

  try
    SyncInternalDataManipulation.Enter;

    Stream := TFileStream.Create(FullFileName, fmOpenWrite);

    Str := 'FSM MACHINE FILE VERSION 1.0'; // Identifikator typu suboru
    Stream.WriteAnsiString(Str);

    Wrd := FILE_CONTENT_STATES; // Identifikator obsahu suboru (stavy automatu)
    Stream.WriteWord(Wrd);

    Int := Length(States); // Pocet stavov automatu
    Stream.WriteBuffer(Int, SizeOf(Int));

    if Length(States) > 0 then
      for I := 0 to High(States) do
      begin
        Stream.WriteBuffer(I, SizeOf(I));
        States[I].MyXY(X, Y);
        Stream.WriteBuffer(X, SizeOf(X));
        Stream.WriteBuffer(Y, SizeOf(Y));
        Bool := States[I].Selected;
        Stream.WriteBuffer(Bool, SizeOf(Bool));
        Bt := Byte(States[I].TypeOfState);
        Stream.WriteByte(Bt);
      end;

    Wrd := FILE_CONTENT_EVENTS; // Identifikator obsahu suboru (udalosti automatu)
    Stream.WriteWord(Wrd);

    Int := Length(StateEvents); // Pocet udalosti spajajucich stavy
    Stream.WriteBuffer(Int, SizeOf(Int));

    if Length(StateEvents) > 0 then
      for I := 0 to High(StateEvents) do
      begin
        Stream.WriteBuffer(I, SizeOf(I));
        X := GetStateIndex(StateEvents[I].Origin);
        Y := GetStateIndex(StateEvents[I].Target);
        Stream.WriteBuffer(X, SizeOf(X));
        Stream.WriteBuffer(Y, SizeOf(Y));
        Str := '';
        for E := #0 to #255 do
          if E in StateEvents[I].UsedSymbols then
            Str := Str + E;
        Stream.WriteAnsiString(Str);
        Db := StateEvents[I].Angle;
        Stream.WriteBuffer(Db, SizeOf(Db));
        Bool := StateEvents[I].Selected;
        Stream.WriteBuffer(Bool, SizeOf(Bool));
        Int := StateEvents[I].Radius;
        Stream.WriteBuffer(Int, SizeOf(Int));
        Bool := StateEvents[I].DisplayLabel;
        Stream.WriteBuffer(Bool, SizeOf(Bool));
        Bt := Byte(StateEvents[I].TypeOfEvent);
        Stream.WriteByte(Bt);
      end;

    Wrd := FILE_CONTENT_MACHINE; // Identifikator obsahu suboru (nastavenie automatu)
    Stream.WriteWord(Wrd);

    Str := '';
    for E := #0 to #255 do
      if E in Alphabet then
        Str := Str + E;
    Stream.WriteAnsiString(Str);
    Stream.WriteBuffer(StartArrowAngle, SizeOf(StartArrowAngle));
    Result := true;
  finally
    SyncInternalDataManipulation.Leave;

    if Stream <> nil then
      Stream.Free;
  end;
end;

function TMachine.LoadFromFile(FullFileName: string; Trans: TApplicationTranslation): Boolean;
var
  Stream: TStream;
  Int, I, Index, E: Integer;
  Str: string;
  Wrd: Word;
  Sections: set of 1..3;
  Alpha, AlphabetLoc: TMachineAlphabet;
  StartArrowAngleLoc: Double;
  Ok: Boolean;
  CacheEvents: array of record
    Index: Integer;
    Origin, Target: Integer;
    Alpha: TMachineAlphabet;
    Angle: Double;
    Selected: Boolean;
    Radius: Integer;
    DisplayLabel: Boolean;
    TypeOf: Byte;
    IsCacheSet: Boolean;
  end;
  CacheStates: array of record
    Index: Integer;
    X,Y: Integer;
    Selected: Boolean;
    TypeOf: Byte;
    IsCacheSet: Boolean;
  end;
begin
  Result := false;
  Sections := [];
  StartArrowAngleLoc := 0;
  Int := 0;

  if not FileExists(FullFileName) then
  begin
    MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_not_exists'], mtError, [mbOK], 0);
    Exit;
  end;

  SyncInternalDataManipulation.Enter;
  try
    Stream := TFileStream.Create(FullFileName, fmOpenRead);

    if Length(States) > 0 then
      for I := 0 to High(States) do
        FreeAndNil(States[I]);
    States := nil;

    if Length(StateEvents) > 0 then
      for I := 0 to High(StateEvents) do
        FreeAndNil(StateEvents[I]);
    StateEvents := nil;

    // Overenie pravosti suboru

    Str := Stream.ReadAnsiString;
    if Str = 'FSM MACHINE FILE VERSION 1.0' then
    begin
      while Stream.Position < Stream.Size do
      begin
        Wrd := Stream.ReadWord; // Nacita identifikator dat
        if Wrd = FILE_CONTENT_MACHINE then
        begin
          // Ochrana pred viacnasobnym citanim sekcie
          if 1 in Sections then
          begin
            MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_unknow_filetype'], mtError, [mbOK], 0);
            Break;
          end;
          Sections := Sections + [1];

          Str := Stream.ReadAnsiString; // Nacita abecedu
          Alpha := []; // Vynulovanie
          for I := 1 to Length(Str) do
            Alpha := Alpha + [Str[I]]; // Rekonstrukcia abecedy
          if Alpha = [] then // Ak je abeceda prazdna, tak nastala chyba
          begin
            MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_data_corruption'], mtError, [mbOK], 0);
            Break;
          end;
          AlphabetLoc := Alpha; // Nastavi abecedu
          Stream.ReadBuffer(StartArrowAngleLoc, SizeOf(StartArrowAngleLoc)); // Nacita rotaciu sipky pociatocneho stavu
        end
        else if Wrd = FILE_CONTENT_STATES then
        begin
          // Ochrana pred viacnasobnym citanim sekcie
          if 2 in Sections then
          begin
            MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_unknow_filetype'], mtError, [mbOK], 0);
            Break;
          end;
          Sections := Sections + [2];

          Stream.ReadBuffer(Int, SizeOf(Int)); // Nacita pocet stavov automatu
          if Int > 0 then
          begin
            SetLength(CacheStates, Int); // Nastavi cache na danu dlzku
            for Index := 0 to High(CacheStates) do
              CacheStates[Index].IsCacheSet := False;

            Ok := true;
            for I := 0 to High(CacheStates) do // Nacita kazdy stav do cache
            begin
              Stream.ReadBuffer(Index, SizeOf(Index)); // Nacita index
              if (Index < 0) or (Index > High(CacheStates)) then
              begin
                MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_data_corruption'], mtError, [mbOK], 0);
                Ok := false;
                Break;
              end;

              if not CacheStates[Index].IsCacheSet then
              begin
                CacheStates[Index].Index := Index; // Zapise sa informacne index
                Stream.ReadBuffer(CacheStates[Index].X, SizeOf(Integer)); // Nacita poziciu X
                Stream.ReadBuffer(CacheStates[Index].Y, SizeOf(Integer)); // Nacita poziciu Y
                Stream.ReadBuffer(CacheStates[Index].Selected, SizeOf(Boolean)); // Nacita stav oznacenia
                Stream.ReadBuffer(CacheStates[Index].TypeOf, SizeOf(Byte)); // Nacita typ stavu
                CacheStates[Index].IsCacheSet := True;
              end
              else
              begin
                MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_data_corruption'], mtError, [mbOK], 0);
                Ok := false;
                Break;
              end;
            end;
            if not Ok then // Ak sa nieco pokazi pri nacitani stavov, tak ukonci nacitanie suboru
              Break;
          end;
        end
        else if Wrd = FILE_CONTENT_EVENTS then
        begin
          // Ochrana pred viacnasobnym citanim sekcie
          if 3 in Sections then
          begin
            MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_unknow_filetype'], mtError, [mbOK], 0);
            Break;
          end;
          Sections := Sections + [3];

          Stream.ReadBuffer(Int, SizeOf(Int)); // Nacita pocet udalosti automatu
          if Int > 0 then
          begin
            SetLength(CacheEvents, Int); // Nastavi cache na danu dlzku
            for Index := 0 to High(CacheEvents) do
              CacheEvents[Index].IsCacheSet := False;

            Ok := true;
            for I := 0 to High(CacheEvents) do // Nacita kazdy stav do cache
            begin
              Stream.ReadBuffer(Index, SizeOf(Index)); // Nacita index
              if (Index < 0) or (Index > High(CacheEvents)) then
              begin
                MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_data_corruption'], mtError, [mbOK], 0);
                Ok := false;
                Break;
              end;

              if not CacheEvents[Index].IsCacheSet then
              begin
                CacheEvents[Index].Index := Index; // Zapise sa informacne index
                Stream.ReadBuffer(CacheEvents[Index].Origin, SizeOf(Integer)); // Nacita sa vychodiskovy stav udalosti
                Stream.ReadBuffer(CacheEvents[Index].Target, SizeOf(Integer)); // Nacita sa cielovy stav udalosti
                Str := Stream.ReadAnsiString; // Nacitaju sa pouzite symboly
                Alpha := [];
                for E := 1 to Length(Str) do
                  Alpha := Alpha + [Str[E]];
                CacheEvents[Index].Alpha := Alpha;
                Stream.ReadBuffer(CacheEvents[Index].Angle, SizeOf(Double)); // Nacita sa rotacia cyklickej udalosti
                Stream.ReadBuffer(CacheEvents[Index].Selected, SizeOf(Boolean)); // Nacita sa vyber udalosti
                Stream.ReadBuffer(CacheEvents[Index].Radius, SizeOf(Integer)); // Nacita radius cyklickej udalosti
                Stream.ReadBuffer(CacheEvents[Index].DisplayLabel, SizeOf(Boolean)); // Nacita zobrazenie symbolov
                Stream.ReadBuffer(CacheEvents[Index].TypeOf, SizeOf(Byte)); // Nacita typ udalosti
                CacheEvents[Index].IsCacheSet := True;
              end
              else
              begin
                MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_data_corruption'], mtError, [mbOK], 0);
                Ok := false;
                Break;
              end;
            end;
            if not Ok then // Ak sa nieco pokazi pri nacitani stavov, tak ukonci nacitanie suboru
              Break;
          end;
        end
        else
        begin
          MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_unknow_filetype'], mtError, [mbOK], 0);
          Break;
        end;
      end;
    end
    else
    begin
      MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_unknow_filetype'], mtError, [mbOK], 0);
    end;
  except
    MessageDlg(Trans['machine_load_from_file_caption'], Trans['machine_load_from_file_error_read'], mtError, [mbOK], 0);
    SyncInternalDataManipulation.Leave;
    if Stream <> nil then
      Stream.Free;
    Exit;
  end;

  if (1 in Sections) and (2 in Sections) and (3 in Sections) then
  begin
    Alphabet := AlphabetLoc;
    StartArrowAngle := StartArrowAngleLoc;

    SetLength(States, Length(CacheStates));
    if Length(CacheStates) > 0 then
      for I := 0 to High(States) do
      begin
        States[CacheStates[I].Index] := TState.Create(CacheStates[I].X, CacheStates[I].Y, AppImage, GraphicLibrary);
        States[CacheStates[I].Index].TypeOfState := TStateType(CacheStates[I].TypeOf mod 4);
      end;

    SetLength(StateEvents, Length(CacheEvents));
    if Length(CacheEvents) > 0 then
      for I := 0 to High(StateEvents) do
      begin
        StateEvents[CacheEvents[I].Index] := TStateEvent.Create(States[CacheEvents[I].Origin], States[CacheEvents[I].Target], AppImage, GraphicLibrary);
        States[CacheEvents[I].Origin].AppendEvent(StateEvents[CacheEvents[I].Index]);
        StateEvents[CacheEvents[I].Index].EventType := TStateEventType(CacheEvents[I].TypeOf mod 2);
        StateEvents[CacheEvents[I].Index].Angle := CacheEvents[I].Angle;
        StateEvents[CacheEvents[I].Index].Radius := CacheEvents[I].Radius;
        StateEvents[CacheEvents[I].Index].Symbols := CacheEvents[I].Alpha;
        StateEvents[CacheEvents[I].Index].DisplayLabel := CacheEvents[I].DisplayLabel;
        StateEvents[CacheEvents[I].Index].RefreshCache;
      end;

    Result := True;
  end;

  SyncInternalDataManipulation.Leave;
  if Stream <> nil then
    Stream.Free;
end;

procedure TMachine.UnboundEvents(Events: array of TStateEvent);
var
  I,E: Integer;
  Found: Boolean;
begin
  SyncInternalDataManipulation.Enter;
  try
    if (Length(StateEvents) > 0) and (Length(Events) > 0) then
      for I := 0 to High(StateEvents) do // Prechadza hlavny zoznam
      begin
        repeat
          Found := false; // Nebola najdena zhoda
          if I <= High(StateEvents) then
            for E := 0 to High(Events) do // Prechadza zoznam zmazavanych
            begin
              if StateEvents[I] = Events[E] then // Zhoda sa nasla
              begin
                StateEvents[I] := StateEvents[High(StateEvents)]; // Nastavy na dany index posledny prvok v poli
                SetLength(StateEvents, Length(StateEvents) - 1); // Skrati pole
                Found := true; // Nastavi najdeny zaznam
                Break; // Vyskoci s cyklu for
              end;
            end;
        until Found = false; // Ak sa nieco naslo, tak sa opat preveri prvok na indexe I
      end;
  finally
    SyncInternalDataManipulation.Leave;
  end;
end;

procedure TMachine.AnimState_Start(var Render: Boolean);
var
  FireTime: TTimeStamp;
  Mx,My: Integer;
begin
  Mx := 0; My := 0;
  Render := true;
  if not Animating then Exit;
  Animation.State_Angle := 0;
  Animation.State_Phase := 0;
  Animation.State_Count := 0;

  if Assigned(AppEventSetOffset) then
  begin
    TState(CurrentAnimatedObject).MyXY(Mx,My);
    AppEventSetOffset(Mx,My);
  end;

  FireTime := TimeStampPlusMicrotime(DateTimeToTimeStamp(Now), 80);
  Renderer.AddAction(Self, @AnimState_Next, FireTime);
end;

procedure TMachine.AnimState_Next(var Render: Boolean);
var
  FireTime: TTimeStamp;
  Mx,My: Integer;
begin
  Mx := 0; My := 0;
  Render := true;
  if not Animating then Exit;
  if GraphicLibrary['Anim_state'] <> nil then
  begin
    Animation.State_Phase := (Animation.State_Phase + 1) mod GraphicLibrary['Anim_state'].Count;
    Animation.State_Angle := (Animation.State_Angle + 360 div GraphicLibrary['Anim_state'].Count) mod 360;
  end;
  Inc(Animation.State_Count);

  if Assigned(AppEventSetOffset) then
  begin
    TState(CurrentAnimatedObject).MyXY(Mx,My);
    AppEventSetOffset(Mx,My);
  end;

  if (GraphicLibrary['Anim_state'] <> nil) and (Animation.State_Count <= GraphicLibrary['Anim_state'].Count) then
  begin
    FireTime := TimeStampPlusMicrotime(DateTimeToTimeStamp(Now), 80);
    Renderer.AddAction(Self, @AnimState_Next, FireTime);
  end
  else
  begin
    if Animation.Sentence = '' then
    begin
      Renderer.Sync(@AnimationEnd);
      //AnimationEnd;
      Exit;
    end
    else if (TState(CurrentAnimatedObject).TypeOfState = mstAcceptAndEndState) or (TState(CurrentAnimatedObject).TypeOfState = mstFailAndEnd) then
    begin
      Renderer.Sync(@AnimationEnd);
      //AnimationEnd;
      Exit;
    end;

    CurrentAnimatedObject := TState(CurrentAnimatedObject).DetermineEvent(Animation.Sentence);
    if CurrentAnimatedObject = nil then
    begin
      Renderer.Sync(@AnimationEnd);
      //AnimationEnd;
    end
    else
    begin
      Animation.Sentence := Copy(Animation.Sentence, 2, Length(Animation.Sentence) - 1);
      AnimEvent_Start(Render);
    end;
  end;
end;

procedure TMachine.AnimEvent_Start(var Render: Boolean);
var
  FireTime: TTimeStamp;
  Mx,My: Integer;
begin
  Mx := 0; My := 0;
  Render := true;
  if not Animating then Exit;
  Animation.Event_Count := 0;
  Animation.Event_Phase := 0;

  if Assigned(AppEventSetOffset) then
  begin
    TStateEvent(CurrentAnimatedObject).Origin.MyXY(Mx,My);
    AppEventSetOffset(Mx,My);
  end;

  FireTime := TimeStampPlusMicrotime(DateTimeToTimeStamp(Now), 80);
  Renderer.AddAction(Self, @AnimEvent_Next, FireTime);
end;

procedure TMachine.AnimEvent_Next(var Render: Boolean);
var
  FireTime: TTimeStamp;
begin
  Render := true;
  if not Animating then Exit;

  if TStateEvent(CurrentAnimatedObject).AnimationCanContinue(Animation.Event_Count) then
  begin
    Inc(Animation.Event_Count);
    if GraphicLibrary['Anim_event'] <> nil then
      Animation.Event_Phase := (Animation.Event_Phase + 1) mod GraphicLibrary['Anim_event'].Count;

    TStateEvent(CurrentAnimatedObject).AnimationSetOffset(Animation.Event_Count, AppEventSetOffset);

    FireTime := TimeStampPlusMicrotime(DateTimeToTimeStamp(Now), 80);
    Renderer.AddAction(Self, @AnimEvent_Next, FireTime);
  end
  else
  begin
    CurrentAnimatedObject := TStateEvent(CurrentAnimatedObject).Target;
    AnimState_Start(Render);
  end;
end;

procedure TMachine.AnimationEnd;
begin
  Animating := false;

  if Assigned(AppOnAnimDone) then
    AppOnAnimDone;

  Renderer.ForceRender;
end;

constructor TMachine.Create(Image: TImage; Alb: TMachineAlphabet; GL: TGraphicLibrary);
begin
  Alphabet := Alb;
  AppImage := Image;
  GraphicLibrary := GL;

  if SyncInternalDataManipulation = nil then
    SyncInternalDataManipulation := TCriticalSection.Create;
end;

constructor TMachine.Create(Image: TImage; GL: TGraphicLibrary);
begin
  Alphabet := [];
  AppImage := Image;
  GraphicLibrary := GL;

  if SyncInternalDataManipulation = nil then
    SyncInternalDataManipulation := TCriticalSection.Create;
end;

procedure TMachine.CreateState(X, Y: Integer; SnapToGrid: Boolean=false);
begin
  if SnapToGrid then
  begin
    X := round(X / 20) * 20;
    Y := round(Y / 20) * 20;
  end;

  if TestCollision(X, Y) then Exit;
  SetLength(States, Length(States) + 1);
  States[High(States)] := TState.Create(X, Y, AppImage, GraphicLibrary);
end;

procedure TMachine.CreateStateEvent(Symbols: string; IsElse: Boolean=false);
var
  NSE: TStateEvent;
  AlphaSymbols: TMachineAlphabet;
  I: Integer;
begin
  if (LastSelState = nil) or (SelState = nil) then Exit; // Skontroluje, ci su naozaj vybrany startovny a cielovy stav

  if LastSelState.IsHereEventTowards(SelState) then Exit; // Skontroluje, ci uz neexistuje stav s tymto smerom

  AlphaSymbols := []; // Vynuluje pouzite znaky
  for I := 1 to Length(Symbols) do
    AlphaSymbols := AlphaSymbols + [Symbols[I]]; // A vytvori novu mnozinu znakov pre udalost

  NSE := LastSelState.NewEvent(SelState); // Vytvori sa nova udalost
  NSE.UsedSymbols := AlphaSymbols; // Nastavia sa jej znaky abecedy
  if IsElse then
    NSE.TypeOfEvent := msetElse
  else
    NSE.TypeOfEvent := msetNormal;

  SetLength(StateEvents, Length(StateEvents) + 1); // Vytvori sa misto pre novu udalost
  StateEvents[High(StateEvents)] := NSE; // Da sa tam udalost
  SelectNone; // A odznacia sa vsetky vybrane stavy
end;

function TMachine.TestEventBetween: Boolean;
begin
  Result := false;
  if (LastSelState = nil) or (SelState = nil) then Exit; // Skontroluje, ci su naozaj vybrany startovny a cielovy stav

  if LastSelState.IsHereEventTowards(SelState) then Exit; // Skontroluje, ci uz neexistuje stav s tymto smerom
  Result := true;
end;

function TMachine.TestElseEventPossibility: Boolean;
begin
  Result := false;

  if (LastSelState = nil) or (SelState = nil) then Exit; // Skontroluje, ci su naozaj vybrany startovny a cielovy stav

  Result := LastSelState.CanHaveElseEvent(Alphabet);
end;

function TMachine.TestMyStateHasElseEvent: Boolean;
var
  I: Integer;
begin
  Result := false;

  if SelStateEvent = nil then Exit;

  SyncInternalDataManipulation.Enter;
  try
    if Length(SelStateEvent.Origin.EventArray) > 0 then
      for I := 0 to High(SelStateEvent.Origin.EventArray) do
        if SelStateEvent.Origin.EventArray[I].TypeOfEvent = msetElse then
        begin
          Result := true;
          Break;
        end;
  finally
    SyncInternalDataManipulation.Leave;
  end;
end;

function TMachine.GetAvailableSymbolsForEvent: TMachineAlphabet;
begin
  Result := Alphabet - LastSelState.GetUsedSymbols(SelStateEvent); // Cela abeceda minus pouzite znaky
end;

procedure TMachine.Render(OfsX, OfsY: Integer);
var
  I, Fx, Fy: Integer;
begin
  Fx := 0; Fy := 0;
  SyncInternalDataManipulation.Enter;
  try
    if Length(States) > 0 then
    begin
      States[0].MyXY(Fx,Fy);
      DrawGraphicImage(AppImage.Canvas, GraphicLibrary, 'Starting_state', 0, Fx + OfsX, Fy + OfsY, StartArrowAngle);
    end;

    if Length(StateEvents) > 0 then
      for I := 0 to High(StateEvents) do
        StateEvents[I].Render(OfsX, OfsY);

    if Length(StateEvents) > 0 then
      for I := 0 to High(StateEvents) do
        StateEvents[I].RenderLabel(OfsX, OfsY, Alphabet);

    if Animating and (CurrentAnimatedObject is TStateEvent) then
    begin
      TStateEvent(CurrentAnimatedObject).RenderAnimatedSpark(OfsX, OfsY, Animation.Event_Phase, Animation.Event_Count);
    end;

    if Length(States) > 0 then
      for I := 0 to High(States) do
      begin
        States[I].Render(OfsX, OfsY, Alphabet);
      end;

    if Animating and (CurrentAnimatedObject is TState) then
    begin
      TState(CurrentAnimatedObject).MyXY(Fx, Fy);
      DrawGraphicImage(AppImage.Canvas, GraphicLibrary, 'Anim_state', Animation.State_Phase, Fx + OfsX, Fy + OfsY, Animation.State_Angle);
    end;
  finally
    SyncInternalDataManipulation.Leave;
  end;
end;

function TMachine.TestSelection(X, Y: Integer):Boolean;
var
  TempSel: TState;
begin
  Result := true;
  TempSel := TestStateSelection(X, Y);
  if TempSel = nil then Result := false;
  LastSelState := SelState;
  SelState := TempSel;
  SelStateEvent := nil;
  if TempSel = nil then
  begin
    SelStateEvent := TestStateEventSelection(X, Y);
  end;
  MarkStateAsSelected(SelState);
  MarkStateEventAsSelected(SelStateEvent);
end;

procedure TMachine.SetXYOfSelState(X, Y: Integer; SnapToGrid: Boolean=false);
var
  I: Integer;
begin
  if SelState = nil then Exit;

  if SnapToGrid then
  begin
    X := round(X / 20) * 20;
    Y := round(Y / 20) * 20;
  end;

  if not TestCollision(X, Y, SelState) then
  begin
    SelState.SetXY(X, Y);

    SyncInternalDataManipulation.Enter;
    try
      if Length(StateEvents) > 0 then
        for I := 0 to High(StateEvents) do
          if (StateEvents[I].Origin = SelState) or (StateEvents[I].Target = SelState) then
            StateEvents[I].RefreshCache;
    finally
      SyncInternalDataManipulation.Leave;
    end;
  end;
end;

procedure TMachine.DeleteSelected;
var
  I: Integer;
  EventsToDelete: TEventArray;
begin
  if SelStateEvent <> nil then
  begin
    UnboundEvents([SelStateEvent]);
    SelStateEvent.Origin.DeleteStateEvent(SelStateEvent);
    SelStateEvent := nil;
    Exit;
  end;

  if SelState <> nil then
  begin
    UnboundEvents(SelState.Events);
    SelState.DeleteAllEvents;

    EventsToDelete := nil;

    SyncInternalDataManipulation.Enter;
    try
    if Length(StateEvents) > 0 then
      for I := 0 to High(StateEvents) do
        if StateEvents[I].Target = SelState then
        begin
          SetLength(EventsToDelete, Length(EventsToDelete) + 1);
          EventsToDelete[High(EventsToDelete)] := StateEvents[I];
        end;
    finally
      SyncInternalDataManipulation.Leave;
    end;

    UnboundEvents(EventsToDelete);
    if Length(EventsToDelete) > 0 then
      for I := 0 to High(EventsToDelete) do
        EventsToDelete[I].Origin.DeleteStateEvent(EventsToDelete[I]);

    SyncInternalDataManipulation.Enter;
    try
      if Length(States) > 0 then
        for I := 0 to High(States) do
          if States[I] = SelState then
          begin
            FreeAndNil(States[I]);
            States[I] := States[High(States)];
            SetLength(States, Length(States) - 1);
            Break;
          end;
    finally
      SyncInternalDataManipulation.Leave;
    end;
    SelState := nil;
    LastSelState := nil;
  end;
end;

procedure TMachine.NumberOfStatesAndEvents(out St, Ev: Integer);
begin
  St := Length(States);
  Ev := Length(StateEvents);
end;

destructor TMachine.Destroy;
var
  I: Integer;
begin
  SyncInternalDataManipulation.Enter;
  if Length(States) > 0 then
    for I := 0 to High(States) do
    begin
      if States[I] <> nil then
        States[I].Free;
    end;
  States := nil;
  if Length(StateEvents) > 0 then
    for I := 0 to High(StateEvents) do
    begin
      if StateEvents[I] <> nil then
        StateEvents[I].Free;
    end;
  StateEvents := nil;
  SyncInternalDataManipulation.Leave;
  Alphabet := [];
  AppImage := nil;

  if SyncInternalDataManipulation <> nil then
    FreeAndNil(SyncInternalDataManipulation);

  inherited Destroy;
end;

end.

