unit renderer;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, Graphics, ExtCtrls, Dialogs, syncobjs;

type
  TPriorityQueueItemAction = procedure(var Return: Boolean) of object;
  TRendererAddAction = procedure(Owner: TObject; Action: TPriorityQueueItemAction; FireAt: TTimeStamp) of object;
  TRendererSync = procedure of object;

  { TPriorityQueueItem }

  TPriorityQueueItem = class
    private
      FireAt: TTimeStamp;
      Owner: TObject;
      Action: TPriorityQueueItemAction;
      Next: TPriorityQueueItem;
      function GetNextItem: TPriorityQueueItem;
      procedure SetNextItem(const AValue: TPriorityQueueItem);
    public
      constructor Create(Fire: TTimeStamp; Own: TObject; Act: TPriorityQueueItemAction; Nxt: TPriorityQueueItem=nil);
      destructor Destroy; override;
      function Test(CurTimeStamp: TTimeStamp): Boolean; // Test ci je cas spustenia mensi alebo rovny casu testu
      function Execute: Boolean; // Vykona akciu tohoto prvku

      property NextItem: TPriorityQueueItem read GetNextItem write SetNextItem; // Cita/nastavuje dalsi prvok
      property GetOwner: TObject read Owner;
  end;

  { TPriorityQueue }

  TPriorityQueue = class
    private
      FirstItem: TPriorityQueueItem;
    public
      constructor Create;
      destructor Destroy; override;
      procedure AddAction(Owner: TObject; Action: TPriorityQueueItemAction; FireAt: TTimeStamp);
      procedure RemoveTickOf(Obj: TObject);
      function TimerTick: Boolean;
  end;

  { TRenderer }

  TRenderer = class(TThread)
    private
      Queue: TPriorityQueue;
      AppObject: TObject;
      QueueOperationSync: TCriticalSection;
      ForceRenderNextTime: Boolean;
      CallScreenResizeSwitch: Boolean;
      OnScreenResize: TRendererSync;

      procedure Render; // Vykreslovanie
    public
      constructor Create(App: TObject); // Konstruktor
      destructor Destroy; override;
      procedure ForceRender; // Nasilne vyrendorovanie obrazu v dalsom ticku casovaca
      procedure AddAction(Owner: TObject; Action: TPriorityQueueItemAction; FireAt: TTimeStamp);
      procedure Sync(Proc: TRendererSync);
      procedure RemoveTickOf(Obj: TObject);
      procedure SetScreenResizeProcedure(OSR: TRendererSync);
      procedure CallScreenResize;
    protected
      procedure Execute; override; // Procedura vlakna
  end;

function TimeStampPlusMicrotime(TS: TTimeStamp; MS: Integer): TTimeStamp;

implementation

uses
  appcontrol;

function TimeStampPlusMicrotime(TS: TTimeStamp; MS: Integer): TTimeStamp;
var
  Days, MilSec: Integer;
begin
  Days := MS div 86400000;
  MilSec := MS mod 86400000;
  Result := TS;
  Result.Time := Result.Time + MilSec;
  Result.Date := Result.Date + Days;
end;

{ TPriorityQueueItem }

function TPriorityQueueItem.GetNextItem: TPriorityQueueItem;
begin
  Result := Next;
end;

procedure TPriorityQueueItem.SetNextItem(const AValue: TPriorityQueueItem);
begin
  Next := AValue;
end;

constructor TPriorityQueueItem.Create(Fire: TTimeStamp; Own: TObject;
  Act: TPriorityQueueItemAction; Nxt: TPriorityQueueItem=nil);
begin
  FireAt := Fire;
  Owner := Own;
  Action := Act;
  Next := Nxt;
end;

destructor TPriorityQueueItem.Destroy;
begin
  FireAt.Date := 0;
  FireAt.Time := 0;
  Owner := nil;
  Action := nil;
  Next := nil;
  inherited Destroy;
end;

function TPriorityQueueItem.Test(CurTimeStamp: TTimeStamp): Boolean;
begin
  Result := false; // Predpokladame, ze test bude neuspesny
  if CurTimeStamp.Date > FireAt.Date then // Ak je vstupny cas (den) vecsi ako cas spustenia (den) tak je test uspesny
    Result := true
  else if CurTimeStamp.Date = FireAt.Date then // Ak sa dni rovnaju
  begin
    if CurTimeStamp.Time >= FireAt.Time then // A vstupny cas (milisekundy) su vacsie alebo rovne casu spustenia (milisekundam) tak je test uspesny
      Result := true;
  end;
end;

function TPriorityQueueItem.Execute: Boolean;
begin
  Result := false;
  if Assigned(Action) then
    Action(Result);
end;

{ TPriorityQueue }

constructor TPriorityQueue.Create;
begin
  FirstItem := nil;
end;

destructor TPriorityQueue.Destroy;
var
  CPQI: TPriorityQueueItem;
begin
  CPQI := FirstItem;
  while CPQI <> nil do
  begin
    FirstItem.Free;
    FirstItem := CPQI.NextItem;
    CPQI := FirstItem;
  end;
  inherited Destroy;
end;

procedure TPriorityQueue.AddAction(Owner: TObject; Action: TPriorityQueueItemAction; FireAt: TTimeStamp);
var
  CPQI, PPQI, TPQI: TPriorityQueueItem;
begin
  if FirstItem = nil then // Ak prvy prvok neexistuje, tak ho vytvori a skonci
    FirstItem := TPriorityQueueItem.Create(FireAt, Owner, Action)
  else
  begin
    PPQI := nil; // Predchadzajuci prvok
    CPQI := FirstItem; // Sucasny prvok
    while CPQI <> nil do // Kym sucasny prvok nie je nil
    begin
      if not CPQI.Test(FireAt) then // Ak nie je pravda, ze cas v testovanom prvku je mensi alebo rovny vkladanemu casu akcie
      begin
        if CPQI = FirstItem then // Ak je sucasny prvok prvym prvkom
        begin
          FirstItem := TPriorityQueueItem.Create(FireAt, Owner, Action, CPQI); // Tak sa vytvori novy prvy prvok a jeho nasledovnik je sucasny prvok
        end
        else
        begin
          TPQI := TPriorityQueueItem.Create(FireAt, Owner, Action, CPQI); // Inak sa vytvori novy prvok s nasledovnikom ako sucasny prvok
          PPQI.NextItem := TPQI; // A tento sa priradi ako nasledovnik predchadzajucemu prvku
        end;
        Exit; // Tu sa skonci vykonavanie proceduri
      end;
      PPQI := CPQI;
      CPQI := CPQI.NextItem;
    end;
    // Ak sa v cykle nevyskocilo z procedury, je v PPQI posledny prvok zoznamu, tak jemu sa prida nasledovnik
    PPQI.NextItem := TPriorityQueueItem.Create(FireAt, Owner, Action);
  end;
end;

procedure TPriorityQueue.RemoveTickOf(Obj: TObject);
var
  CPQI, PPQI: TPriorityQueueItem;
begin
  if FirstItem <> nil then // Iba ak prvy objekt existuje
  begin
    CPQI := FirstItem; // Sucasny prvok
    PPQI := nil; // Predchodzi prvok
    while CPQI <> nil do // Kym sucasny prvok existuje
    begin
      if CPQI.GetOwner = Obj then // Ak sme nasli zhodu ...
      begin
        if PPQI = nil then // ... a predchodca neexistuje
        begin
          FirstItem := CPQI.NextItem; // Nastav novy prvy prvok
          CPQI.Free; // Vyhod sucasny
          CPQI := FirstItem; // Nastav ako sucasny prvok prvy prvok v zozname
          Continue; // Pokracuj dalsou iteraciou
        end
        else // ... a predchodca existuje
        begin
          PPQI.NextItem := CPQI.NextItem; // Nastav predchodcovi nasledovnika na nasledovnika sucasneho prvku
          CPQI.Free; // Vyhod sucasny
          CPQI := PPQI.NextItem; // Sucasny prvok je nasledovnik predchodcu (ten sa nemeni)
          Continue; // Pokracuj dalsou iteraciou
        end;
      end;
      PPQI := CPQI; // Posuvanie prvkov (ked sa nic nenajde)
      CPQI := CPQI.NextItem;
    end;
  end;
end;

function TPriorityQueue.TimerTick: Boolean;
var
  CurTime: TTimeStamp;
  TPQI: TPriorityQueueItem;
begin
  Result := false; // Nastavime na false, predpokladame, ze prvy prvok nebude mat cas spustenia starsi ako sucasny cas
  if FirstItem <> nil then // Iba ak prvy prvok existuje
  begin
    CurTime := DateTimeToTimeStamp(Now); // Zistime sucasny cas

    if FirstItem.Test(CurTime) then // Ak je cas spustenia prveho prvku starsi ako cas sucasny
    begin
      Result := FirstItem.Execute; // Vykoname prvy prvok a ziskame vystup

      TPQI := FirstItem.NextItem; // Zapamatame nasledujuci prvok
      FirstItem.Free; // Prvy vyhodime
      FirstItem := TPQI; // A ako prvy nastavime jeho povodne nasledujuci

      Result := Result or TimerTick; // Vykoname test na dalsom prvku
    end;
  end;
end;

{ TRenderer }

procedure TRenderer.Render;
begin
  TApplicationControl(AppObject).RenderAll;
end;

constructor TRenderer.Create(App: TObject);
begin
  QueueOperationSync := TCriticalSection.Create;
  AppObject := App;
  Queue := TPriorityQueue.Create;
  ForceRenderNextTime := False;
  inherited Create(true);
end;

destructor TRenderer.Destroy;
begin
  Queue.Free;
  QueueOperationSync.Free;
  inherited Destroy;
end;

procedure TRenderer.ForceRender;
begin
  ForceRenderNextTime := true;
end;

procedure TRenderer.AddAction(Owner: TObject; Action: TPriorityQueueItemAction;
  FireAt: TTimeStamp);
begin
  QueueOperationSync.Enter;
  try
    Queue.AddAction(Owner, Action, FireAt);
  finally
    QueueOperationSync.Leave;
  end;
end;

procedure TRenderer.Sync(Proc: TRendererSync);
begin
  if Assigned(Proc) then
    Synchronize(Proc);
end;

procedure TRenderer.RemoveTickOf(Obj: TObject);
begin
  QueueOperationSync.Enter;
  try
    Queue.RemoveTickOf(Obj);
  finally
    QueueOperationSync.Leave;
  end;
end;

procedure TRenderer.SetScreenResizeProcedure(OSR: TRendererSync);
begin
  OnScreenResize := OSR;
end;

procedure TRenderer.CallScreenResize;
begin
  CallScreenResizeSwitch := true;
end;

procedure TRenderer.Execute;
var
  Test: Boolean;
begin
  while not Terminated do
  begin
    if CallScreenResizeSwitch and Assigned(OnScreenResize) then
      Synchronize(OnScreenResize);
    Test := false;
    QueueOperationSync.Enter;
    try
      Test := Queue.TimerTick; // Kontrola prvko, vrati informaciu ci renderovat alebo nie
    finally
      QueueOperationSync.Leave;
    end;
    if ForceRenderNextTime or Test then
      Synchronize(@Render); // Vykreslovanie
    ForceRenderNextTime := false;
    CallScreenResizeSwitch := false;
    Sleep(10); // Cakaj stotinu sekundy
  end;
end;

end.

