unit translation;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, math, Dialogs;

type
  TApplicationTranslationRecord = record
    LangLabel: string;
    LangText: string;
  end;

  { TApplicationTranslation }

  TApplicationTranslation = class
    private
      HashTable: array of TApplicationTranslationRecord;
      HashTableSize: Cardinal;
      HashCorrectionStep: Cardinal;
      Path: string;

      procedure SetHashTableSize(NewSize: Cardinal); // Zmeni velkost hashovacej tabulky
      function CountLines(var F: TextFile): Cardinal; // Vrati pocet riadkov v otvorenom subore

      procedure DetermineTableSizeAndStep(CountOfLines: Cardinal); // Vypocita optimalnu velkost tabulky pre hashovanie a korekcny krok
      function SplitByEquationSign(Line: string): TApplicationTranslationRecord; // Rozdeli riadok podla znaku pre rovnost

      function GetLabelPosition(LangLabel: string; Strict: Boolean=false): Cardinal;
      procedure InsertToTable(Rec: TApplicationTranslationRecord); // Vlozi zaznam do tabulky
    public
      constructor Create(AppPath: string); // Vytvori objekt prekladaca
      destructor Destroy; override; // Zrusi objekt prekladaca

      function LoadLanguage(LangTag: string): Boolean;

      function GetLanguageText(LangLabel: string): string;

      property LangText[LangLabel: string]: string read GetLanguageText; default;
  end;

implementation

{ TApplicationTranslation }

procedure TApplicationTranslation.SetHashTableSize(NewSize: Cardinal);
begin
  HashTableSize := NewSize;

  SetLength(HashTable, HashTableSize);
end;

function TApplicationTranslation.CountLines(var F: TextFile): Cardinal;
var
  Line: string;
begin
  Result := 0;
  Reset(F);
  while not SeekEOF(F) do
  begin
    ReadLn(F, Line);
    Line := Trim(Line);
    if Line[1] <> '#' then
      Inc(Result);
  end;
  Reset(F);
end;

procedure TApplicationTranslation.DetermineTableSizeAndStep(CountOfLines: Cardinal);
var
  TableSize: Cardinal;
  Exponent: SmallInt;
begin
  TableSize := Round(CountOfLines * 2);

  for Exponent := 1 to 32 do
  begin
    if TableSize <= power(2,Exponent) then
      Break;
  end;

  TableSize := floor(power(2,Exponent));

  HashCorrectionStep := TableSize div 16;
  if HashCorrectionStep = 0 then
    HashCorrectionStep := 1;

  SetHashTableSize(TableSize);
end;

function TApplicationTranslation.SplitByEquationSign(Line: string): TApplicationTranslationRecord;
var
  EqPos: Cardinal;
begin
  EqPos := Pos('=', Line);
  Result.LangLabel := Trim(Copy(Line, 1, EqPos - 1));
  Result.LangText := Trim(Copy(Line, EqPos + 1, MaxInt));
end;

function TApplicationTranslation.GetLabelPosition(LangLabel: string; Strict: Boolean=false): Cardinal;
var
  Sum, I, Cycles: Cardinal;
begin
  if HashTableSize = 0 then
  begin
    Result := 0;
    Exit;
  end;
  Sum := 0;

  for I := Length(LangLabel) downto 1 do
  begin
    Sum := Sum + I * Ord(LangLabel[I]);
  end;

  Result := Sum mod HashTableSize;
  if ((HashTable[Result].LangLabel = '') and (Strict = false)) or (HashTable[Result].LangLabel = LangLabel) then
    Exit;

  Cycles := 0;

  while not (((HashTable[Result].LangLabel = '') and (Strict = false)) or (HashTable[Result].LangLabel = LangLabel)) do
  begin
    Result := (Result + HashCorrectionStep) mod HashTableSize;
    Inc(Cycles);
    if Cycles >= HashTableSize then
    begin
      Result := 0;
      Exit;
    end;
  end;
end;

procedure TApplicationTranslation.InsertToTable(Rec: TApplicationTranslationRecord);
begin
  if HashTableSize = 0 then Exit;
  Rec.LangText := StringReplace(Rec.LangText, '\n', #13#10, [rfReplaceAll]);
  HashTable[GetLabelPosition(Rec.LangLabel)] := Rec;
end;

constructor TApplicationTranslation.Create(AppPath: string);
begin
  Path := AppPath;
  HashTable := nil;
  HashTableSize := 0;
end;

destructor TApplicationTranslation.Destroy;
begin
  SetLength(HashTable, 0);
  HashTable := nil;

  inherited Destroy;
end;

function TApplicationTranslation.LoadLanguage(LangTag: string): Boolean;
var
  TF: TextFile;
  Lines: Cardinal;
  Line: string;
  Rec: TApplicationTranslationRecord;
begin
  Result := false;

  if not FileExists(Path + '\language\' + LangTag + '.lng') then
  begin
    if LangTag = 'default' then
    begin
      Exit;
    end;

    Result := LoadLanguage('default');
    Exit;
  end;

  try
    AssignFile(TF, Path + '\language\' + LangTag + '.lng');
    Lines := CountLines(TF);

    HashTable := Nil;

    DetermineTableSizeAndStep(Lines);

    while not SeekEOF(TF) do
    begin
      ReadLn(TF, Line);
      Line := Trim(Line);
      if Line[1] <> '#' then
      begin
        Rec := SplitByEquationSign(Line);
        InsertToTable(Rec);
      end;
    end;
  finally
    CloseFile(TF);
    Result := true;
  end;
end;

function TApplicationTranslation.GetLanguageText(LangLabel: string): string;
var
  Position: Cardinal;
begin
  if HashTableSize = 0 then
  begin
    Result := LangLabel;
    Exit;
  end;
  Result := '';
  Position := GetLabelPosition(LangLabel, true);
  if HashTable[Position].LangLabel = LangLabel then
    Result := HashTable[Position].LangText
  else
    Result := LangLabel;
end;

end.

